"""
=================
fit_melting_paths
=================

Find best-fitting melting paths.
"""

import warnings

import numpy as np
import pandas as pd
from scipy.optimize import minimize_scalar
import shapely.geometry as shp

import pyMelt as m

def find_max_potential_temperature(mantle):
    """
    Find the maximum valid potential temperature for a given pyMelt mantle
    object.
    
    Solidi described by a parabolic function, as in Katz et al. (2003, 
    G-cubed), have turning points at high temperatures/pressures. As a result,
    adiabatic geotherms above some potential temperature no longer intersect
    the solidus, which is a problem for the fitting functions in meltPT. This
    function finds the maximum potential temperature that can be safely used
    with a given mantle object. 
    
    Parameters
    ----------
    mantle : instance of pyMelt.mantle_class.mantle
        The mantle object to be used to calculate the melting path.
    """
    
    # Loop over lithologies, find potential temperature corresponing to solidus
    # turning point.
    Tp = []
    for lith in mantle.lithologies:
        
        # Turning point pressure & temperature.
        # Obtained by differentiating solidus and setting gradient to zero.
        P_tp = -lith.parameters['A2'] / (2.*lith.parameters['A3'])
        T_tp = lith.TSolidus(P_tp)
        
        # Corresponding potential temperature.
        Tp.append( 
            (T_tp+273.) / 
            (np.exp(P_tp * lith.alphas / (lith.rhos*lith.CP)))-273.
            )

    # True maximum will be a bit higher.
    # Increment Tp by one degree at a time until intersection can no longer
    # be found.
    Tp = np.ceil( min(Tp) )
    while ~np.isnan(mantle.solidusIntersection(Tp)):
        Tp += 1.
    
    return Tp


def melt_fraction_to_pressure_temperature(F, path):
    """
    Find pressure and temperature corresponding to specified melt fraction
    for a specified melting path.
    
    Parameters
    ----------
    F : float
        The melt fraction.
    path : instance of pyMelt.meltingcolumn_classes.meltingColumn
        The melt path.
        Can be instance of any class containing arrays of melt fraction (F),
        pressure (P) and temperature (T).
        
    Returns
    -------
    P : float
        Pressure corresponding to specified melt fraction.
    T : float 
        Temperature corresponding to specified melt fraction.
    """
    P = np.interp(F, path.F, path.P)
    T = np.interp(F, path.F, path.T)
    return P, T
    
def compute_sample_melt_fraction_misfit(F, df, path):
    """
    Compute misfit between sample and point along given melting path,
    specified by melt path.
    
    Parameters
    ----------
    F : float
        The melt fraction with which to compare the pressure/temperature
        point.
    df : pandas dataframe
        Dataframe containing sample pressure and temperature estimates.
    path : instance of pyMelt.meltingcolumn_classes.meltingColumn
        The melt path.
        Can be instance of any class containing arrays of melt fraction (F),
        pressure (P) and temperature (T).
    P_err : float, optional
        Uncertainty in pressure observation(s). If NaN, P_err is assumed to 
        be 10% of P.
    T_err : float, optional
        Uncertainty in temperature observation(s). If NaN, T_err is assumed to 
        be 10% of T.

    Returns
    -------
    misfit : float
        Misfit between observed pressure and temperature and point along
        melt path. Defined as distance in normalised pressure-temperature
        space. Pressure and temperature are normalised by their respective
        uncertainties.
    """
    
    if np.isnan(df['P_err']):
        df['P_err'] = df['P'] * 0.1
    if np.isnan(df['T_err']):
        df['T_err'] = df['T'] * 0.1    
        
    model_P, model_T = melt_fraction_to_pressure_temperature(F, path)

    misfit = np.sqrt( 
        float(((df['P']-model_P)/df['P_err'])**2. + ((df['T']-model_T)/df['T_err'])**2.)
        )
    return misfit
    
def find_sample_melt_fraction(df, path):
    """
    Find best-fitting melt fraction for sample along given melting path.
    
    Uses scipy's minimize_scalar function to find melt fraction that minimizes 
    misfit between estimated pressure and temperature and the melt path.
    Options used are:
        method: "bounded"
        bounds:  min --> 0
                 max --> Maximum melt fraction in "path"
        bracket: min --> 0
                 max --> Maximum melt fraction in "path"
        
    Parameters
    ----------
    df : pandas dataframe
        Dataframe containing sample pressure and temperature estimates.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
    path : instance of pyMelt.meltingcolumn_classes.meltingColumn
        The melting path.
        Can be instance of any class containing arrays of melt fraction (F),
        pressure (P) and temperature (T).
        
    Returns
    -------
    out : dict
        Various properties of the result:
            F : float 
                The best-fitting melt fraction.
            P : float
                Pressure on path corresponding to best-fitting melt fraction.
            T : float
                Temperature on path corresponding to best-fitting melt fraction.
            misfit : float
                Distance between sample and closest point on melting path.
        If sample pressure or temperature are nan, all returned values are nan.
    """
    if not df['Fit_Tp']:
        out = {'F': np.nan, 'P': np.nan, 'T': np.nan, 'misfit': np.nan}
    else:
        fit = minimize_scalar(
            compute_sample_melt_fraction_misfit, 
            bounds=(0.,max(path.F)), 
            bracket=(0.,max(path.F)), 
            args=(df, path), 
            method="bounded")
        P, T = melt_fraction_to_pressure_temperature(fit.x, path)
        out = {'F': fit.x, 'P': P, 'T': T, 'misfit': fit.fun}
    return out

def compute_sample_potential_temperature_misfit(Tp, df, mantle):
    """
    Compute a melting path for a given potential temperature then find misfit
    between it and a sample pressure-temperature estimate.
    
    First uses mantle.adiabaticMelt() to compute melt path for specified
    potential temperature. Then finds closest melt fraction 
    
    Parameters
    ----------
    Tp : float
        The potential temperature to be used.
    df : pandas dataframe
        Dataframe containing sample pressure and temperature estimates.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
    mantle : instance of pyMelt.mantle_class.mantle
        The mantle object to be used to calculate the melting path.
    
    Returns
    -------
    misfit : float
        Distance between sample pressure-temperature estimate and its nearest
        point on the calculated melting path.
    """
    path = mantle.adiabaticMelt(
        Tp, 
        Pstart=max(mantle.solidusIntersection(Tp))+0.01,
        dP=-0.01)
    misfit = find_sample_melt_fraction(df, path)['misfit']
    return misfit

def find_sample_potential_temperature(df, mantle):
    """
    Find best-fitting potential temperature for a sample pressure-temperature
    estimate.
    
    Uses scipy's minimize_scalar function to find potential temperature that
    minimizes misfit between estimated pressure and temperature and the
    corresponding  melt path. Options used are:
        method: "bounded"
        bounds:  min --> intersection of solidus with surface
                 max --> max Tp for melting model
        bracket: min --> intersection of solidus with surface
                 max --> max Tp for melting model
                 
    Parameters
    ----------
    df : pandas dataframe
        Dataframe containing sample pressure and temperature estimates.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
    mantle : instance of pyMelt.mantle_class.mantle
        The mantle object to be used to calculate the melting paths.
        
    Returns
    -------
    out : dict
        Various properties of the result:
            F : float 
                The melt fraction along the best-fitting melting path.
            P : float
                The pressure along the best-fitting melting path.
            T : float
                The temperature along the best-fitting melting path.
            misfit : float
                Distance between sample and closest point on melting path.
            Tp : float
                The best-fitting potential temperature.
            path : instance of pyMelt.meltingcolumn_classes.meltingColumn
                The best-fitting melting path.
        If sample pressure or temperature are nan, all returned values are nan.
    """
    if not df['Fit_Tp']:
        out = {
            'F': np.nan, 'P': np.nan, 'T': np.nan, 
            'misfit': np.nan, 'Tp': np.nan, 'path': np.nan}
    else:
        max_Tp = find_max_potential_temperature(mantle)
        Tp_fit = minimize_scalar(
            compute_sample_potential_temperature_misfit, 
            bracket=(min([lith.TSolidus(0.) for lith in mantle.lithologies]),max_Tp), 
            bounds=(min([lith.TSolidus(0.) for lith in mantle.lithologies]),max_Tp), 
            args=(df,mantle), 
            method="bounded")
        path = mantle.adiabaticMelt(
            Tp_fit.x, 
            Pstart=max(mantle.solidusIntersection(Tp_fit.x))+0.01, 
            dP=-0.01)
        if Tp_fit.success:
            out = find_sample_melt_fraction(df, path)
            out['Tp'] = Tp_fit.x 
            out['path'] = path
        else:
            out = {
                'F': np.nan, 'P': np.nan, 'T': np.nan, 
                'misfit': np.nan, 'Tp': np.nan, 'path': np.nan}
    return out
        
def compute_suite_potential_temperature_misfit(Tp, df, mantle):
    """
    Compute a melt path for a specified potential temperature then calculate
    the misfit between it and one or more pressure-temperature estimates.
    
    Parameters
    ----------
    Tp : float
        The potential temperature to be used.
    df : pandas dataframe
        Dataframe containing sample pressure and temperature estimates.
        Can contain any number of samples.
    mantle : instance of pyMelt.mantle_class.mantle
        The mantle object to be used to calculate the melting paths.
    
    Returns
    -------
    misfit : float
        The average distance between sample pressure-temperature estimates
        and their nearest points on the calculated melting path.
    """
    path = mantle.adiabaticMelt(
        Tp,
        Pstart=max(mantle.solidusIntersection(Tp))+0.01, 
        dP=-0.01)
    melt_fraction_fits = df.apply(
        find_sample_melt_fraction, 
        axis=1, 
        result_type="expand", 
        args=(path,))
    misfit = np.nanmean(melt_fraction_fits['misfit'])
    return misfit

def find_bounding_potential_temperature(points, starting_temperature, mantle, lower=False, threshold=(2./3.)):
    """
    Find either upper or lower bound on best-fitting potential temperature for
    suite of pressure-temperature estimates.
    
    Works by computing melting paths progressively further away from the best-
    fitting path, until a threshold number of points lie between the two paths.
    
    Parameters
    ----------
    points : list of shapely.geometry.point.Point objects
        The points to be bounded.
    starting_temperature : float
        The best-fitting potential temperature for the suite.
    mantle : instance of pyMelt.mantle_class.mantle
        The mantle object to be used to calculate the melting paths.
    lower : bool, optional
        Specify whether an upper or lower bound is to be found.
    threshold : float
        The threshold fraction of points to be lie between the best-fitting
        and bounding temperature melting paths.
    
    Returns
    -------
    bounding_temperature : float
        The estimated bounding temperature.
    bounding_path : instance of pyMelt.meltingcolumn_classes.meltingColumn
        The melting path corresonding to the bounding potential temperature.
    """
    
    # First check whether an upper or lower bound is to be found and define
    # initial guess and the adjustment direction accordingly.
    # If searching for an upper bound, initial guess given by rounding up
    # best-fitting temperature to nearest whole degree and increment is
    # positive. Otherwise, we round down and the increment is negative.
    if not lower:
        bounding_temperature = np.ceil(starting_temperature)
        adjustment = 1.
    else:
        bounding_temperature = np.floor(starting_temperature)
        adjustment = -1.

    # Initialise array specifying which points are inside (1) vs. outside (0)
    # the bounding melting paths.
    inside = np.zeros(len(points))

    # Compute melting path corresponding to best-fitting potential temperature
    # Use largest of solidus pressure or maximum pressure in points as the
    # starting pressure.
    max_P = max([p.coords.xy[1][0] for p in points])
    main_path = mantle.adiabaticMelt(
        starting_temperature,
        Pstart=max(max(mantle.solidusIntersection(starting_temperature))+0.01, max_P),
        dP=-0.01)

    # Bounding potential temperatures for given mantle object
    max_Tp = find_max_potential_temperature(mantle)
    min_Tp = min([lith.TSolidus(0.) for lith in mantle.lithologies])
        
    # Start incrementally expanding bounds.
    while True:
        
        # Check bounds have not been exceeded
        if bounding_temperature >= max_Tp or bounding_temperature <= min_Tp:
            bounding_temperature = np.nan
            bounding_path = None
            break
        
        # Compute melting bath corresponding to bounding potential temperature.
        bounding_path = mantle.adiabaticMelt(
            bounding_temperature,
            Pstart=max(max(mantle.solidusIntersection(bounding_temperature))+0.01, max_P), 
            dP=-0.01)
        
        # Create bounding polygon.
        bounds = np.vstack(( 
            np.column_stack(( main_path.T, main_path.P )),
            np.column_stack(( bounding_path.T[::-1], bounding_path.P[::-1] ))
            ))
        poly = shp.polygon.Polygon(bounds)
        
        # Check which points lie inside polygon.
        for i,p in enumerate(points):
            if poly.contains(p):
                inside[i] = 1.
        
        # Stop if we have reached threshold, otherwise increment and continue.
        if sum(inside) / len(points) > threshold:
            break
        else:
            bounding_temperature += adjustment
            
    return bounding_temperature, bounding_path



"""
======
meltPT
======

A Python package for whole-rock major-element themormobarometric analyses of basaltic (mafic) rocks.

"""


from .parse import *
from .backtrack_compositions import *
from .thermobarometry import *
from .fit_melting_paths import *
from .suite import *


"""
=====
parse
=====

Read data from a csv file.
"""

import warnings

import numpy as np
import pandas as pd

def parse_csv(
        infile, Ce_to_H2O=0., src_FeIII_totFe=0., src_Fo=0.9, min_MgO=0., 
        param_co2=False):
    """
    Read a csv and return a dataframe after some processing.
    
    Processing steps are:
    - check SiO2, MgO, and FexOx are specified; if not will crash
    - check other major elements are specified; if not will be set to zero
    - try to set some values via trace elements
    - redistribute Fe according to src_FeIII_totFe
    - add source forsterite numbers if not already specified
    - if desired, estimate CO2 from SiO2
    - normalise major elements to 100%
    - reject samples with MgO less than some threshold.
      
    Parameters
    ----------
    infile : str
        Path to a csv containing data to be read.
    Ce_to_H2O : float
        Ratio of Ce to H2O in mantle source.
    src_FeIII_totFe : float
        Ratio of Fe3+ to total Fe in the mantle source.
    src_Fo : float
        Forsterite number in the mantle source.
    min_MgO : float, optional
        Minimum amount of MgO in sample to be accepted.
    read_as_primary : bool
        If true, data from input_csv is assumed to be primary
        and backtracking is avoided.
    param_co2 : bool
        If true, CO2 is calculated from SiO2 concentration.
        
    Returns
    -------
    df : pandas dataframe
        Dataframe containing processed data.
    """

    # Read in file
    df = pd.read_csv(infile, delimiter=",")
    
    # Check for compulsory columns
    compulsory_cols = ['SiO2', 'MgO']
    for col in compulsory_cols:
        if col not in df.columns:
            raise Exception("Input csv must contain a %s column." % col)
    if ("FeO" not in df.columns and 
        "Fe2O3" not in df.columns and 
        "FeO_tot" not in df.columns):
        raise Exception("Input csv must contain one of FeO, Fe2O3, FeO_tot columns.")

    # Replace empties and NaNs with zeros
    df = df.replace(r'^\s*$', 0., regex=True)
    df = df.replace(np.nan, 0., regex=True)

    # If major element columns do not exist in the input make them and
    # give them zeros for every row.
    major_columns = ['SiO2', 'TiO2', 'Al2O3', 'FeO', 'Fe2O3', 'MgO', 'CaO', 
                     'Na2O', 'K2O', 'MnO', 'Cr2O3', 'P2O5', 'NiO', 'CoO',
                     'H2O', 'CO2']
    for col in major_columns + ["FeO_tot"]:
        if col not in df.columns:
            message = "Input csv does not contain a %s column: we will try to fill it for you, or set it to zero." % col
            warnings.warn(message)
            df[col] = 0.
            
    # If not given compute some major oxides from trace element concentrations
    # df.loc[df['P2O5'] == 0, 'P2O5'] = df['P'] * 141.942524 / 2. / 10000.
    try:
        df.loc[df['NiO']==0., 'NiO'] = df['Ni'] * 1.2725 / 10000.
    except KeyError:
        pass
    try:
        df.loc[df['CoO']==0., 'CoO'] = df['Co'] * 1.2715 / 10000.
    except KeyError:
        pass
    try:
        df.loc[df['Cr2O3']==0., 'Cr2O3'] = df['Cr'] * 1.4616 / 10000.
    except KeyError:
        pass
        
    # Calculate H2O value if H2O value is zero
    # parameterizes water by converting Ce to H20
    if (df['H2O'] == 0.).any():
        try:
            df.loc[df['H2O']==0., 'H2O'] = df['Ce'] * Ce_to_H2O / 10000.
        except KeyError:
            message = "Some sample's H2O still zero after parameterization with Ce."
            warnings.warn(message)

    # Add chosen FeIII_totFe value if none are given
    if 'src_FeIII_totFe' not in df.columns:
        df['src_FeIII_totFe'] = src_FeIII_totFe
    df.loc[df['src_FeIII_totFe'] == 0, 'src_FeIII_totFe'] = src_FeIII_totFe
    df.loc[df['FeO_tot']==0., 'FeO_tot'] = df['FeO'] + (df['Fe2O3'] * (74.84 * 2.) / 159.69)
    
    # Calculate FeO and Fe2O3 based on FeIII_totFe    
    # Only applicable if FeO and Fe2O3 not already stated.
    df.loc[(df['FeO']==0.) | (df['Fe2O3']==0.), 'FeO'] = df['FeO_tot'] * (1. - df['src_FeIII_totFe'])
    df.loc[(df['FeO']==0.) | (df['Fe2O3']==0.), 'Fe2O3'] = df['FeO_tot'] * df['src_FeIII_totFe'] * 159.69 / 71.84 / 2.      

    # Add chosen Fo# if none are given
    if 'src_Fo' not in df.columns:
        df['src_Fo'] = src_Fo
    df.loc[df['src_Fo'] == 0, 'src_Fo'] = src_Fo

    # Calculate CO2 value if desired and CO2 value is zero
    # parameterize CO2 using Equation 8 of Sun & Dasgupta, 2020    
    if param_co2:
        df.loc[df['CO2'] == 0, 'CO2'] = 43.77 - 0.9 * df['SiO2']

    # Calculate total & normalise
    df['Total'] = df[major_columns].sum(axis=1)
    for col in major_columns:
        df[col] = df[col] / df['Total'] * 100.
    df['Total'] = df[major_columns].sum(axis=1)

    # Filter dataset to only include usable samples
    # Remove lines with MgO < 8.5 wt%
    df = df.loc[(df['MgO'] > min_MgO)]

    return df


"""
===============
The Suite class
===============

Process suites of samples.
"""

from .parse import *
from .backtrack_compositions import *
from .thermobarometry import *
from .fit_melting_paths import *

MAJOR_OXIDES = [
    'SiO2','Al2O3','FeO','Fe2O3','MgO','CaO','Na2O','K2O','TiO2','MnO',
    'Cr2O3', 'P2O5', 'NiO', 'CoO', 'H2O', 'CO2']

class Suite:
    """
    Store and process compositions of basaltic rocks.
    
    Includes methods to find primary compositions (i.e. correcting for the
    crystallisation of olivine; Lee et al., 2009, EPSL), compute equilibration
    pressures and temperatures (Plank & Forsyth, 2016, G-cubed), and fit
    melting paths to those pressure-temperature estimates.
    
    Parameters
    ----------
    input_csv : str
        Path to a csv containing data to be read.
    Ce_to_H2O : float
        Ratio of Ce to H2O in mantle source.
    src_FeIII_totFe : float
        Ratio of Fe3+ to total Fe in the mantle source.
    min_MgO : float, optional
        Minimum amount of MgO in sample to be accepted.
    read_as_primary : bool
        If true, data from input_csv is assumed to be primary
        and backtracking is avoided.
    param_co2 : bool
        If true, CO2 is calculated from SiO2 concentration.
    read_PT : bool
        If true, pressures and temperatures are not calculated.
        
    Attributes
    ----------
    data : pandas dataframe
        The raw data from the provided csv.
    primary : pandas dataframe or NoneType
        If backtrack_compositions has been run, or read_as_primary is set 
        to true, will contain the estimated primary compositions in 
        various forms.
    PT : pandas dataframe or NoneType
        If compute_pressure_temperature has been run, will contain the
        esimated equilibration pressures and temperatures.
    PT_to_fit : pandas dataframe or NoneType
        If any melt-path fitting has been run, will contain pressure and
        temperature estimates of samples that have been selected for fitting.
    individual_melt_fractions : pandas dataframe or NoneType
        If find_individual_melt_fractions, will contain results of fitting
        suite of pressure-temperature estimates to a specified melt path.
    individual_potential_temperatures : pandas dataframe or NoneType
        If find_individual_potential_temperatures has been run, will contain
        results of fitting melting paths to each individual pressure-
        temperature estimate.
    suite_melt_fractions : pandas dataframe or NoneType
        If find_suite_potential_temperature has been run, will contain closest
        points on best-fitting melting path for each pressure-temperature
        estimate.
    potential_temperature : float or NoneType
        If find_suite_potential_temperature has been run, will be the best-
        fitting potential temperature for the suite of pressure-temperature
        estimates.
    upper_potential_temperature : float or NoneType
        If find_suite_potential_temperature has been run, with find_bounds,
        will be the upper bound on potential temperature for the suite of
        pressure-temperature estimates.
    lower_potential_temperature : float or NoneType
        If find_suite_potential_temperature has been run, with find_bounds,
        will be the lower bound on potential temperature for the suite of
        pressure-temperature estimates.
    path : instance of pyMelt.meltingcolumn_classes.meltingColumn
        If find_suite_potential_temperature has been run, will be the best-
        fitting melting path.
    upper_path : instance of pyMelt.meltingcolumn_classes.meltingColumn
        If find_suite_potential_temperature has been run, with find_bounds,
        will be the upper-bounding melting path.
    lower_path : instance of pyMelt.meltingcolumn_classes.meltingColumn
        If find_suite_potential_temperature has been run, with find_bounds,
        will be the lower-bounding melting path.
    """

    def __init__(
            self, input_csv, Ce_to_H2O=0., src_FeIII_totFe=0., src_Fo=0.9,
            min_MgO=0., read_as_primary=False, param_co2=False, read_PT=False):
        self.data = parse_csv(
            input_csv,
            Ce_to_H2O=Ce_to_H2O,
            src_FeIII_totFe=src_FeIII_totFe,
            src_Fo=src_Fo,
            min_MgO=min_MgO,
            param_co2=param_co2)
        self.PT_to_fit = None
        self.individual_melt_fractions = None
        self.individual_potential_temperatures = None
        self.suite_melt_fractions = None
        self.potential_temperature = None
        self.upper_potential_temperature = None
        self.lower_potential_temperature = None
        self.path = None
        self.upper_path = None
        self.lower_path = None
        
        if read_as_primary:
            self.primary = self.data.filter(items=MAJOR_OXIDES, axis=1)
            primary_labels = {x: x + "_primary_wt" for x in MAJOR_OXIDES}
            self.primary = self.primary.rename(primary_labels, axis=1)
        else:
            self.primary = None
            
        if read_PT:
            self.PT = self.data.filter(items=['P', 'T'], axis=1)
            self.data = self.data.drop(labels=['P', 'T'], axis=1)
        else:
            self.PT = None
            self.data = self.data.drop(labels=['P', 'T'], axis=1, errors='ignore')

    def backtrack_compositions(self, backtracker):
        """
        Backtrack compositions for entire suite.
        
        Applies backtracker.backtrack_sample_composition to the "data" property.
        Result is saved in the "primary" property.
        
        Parameters
        ----------
        backtracker : class instance
            Instance of a backtracker class. Must have a
            backtrack_sample_composition method that returns a dict containing
            backtracked major element compositions.        
        """
        self.primary = self.data.apply(
            backtracker.backtrack_sample_composition,
            axis=1,
            result_type="expand"
            )

    def compute_pressure_temperature(self, method="PF16", min_SiO2=0.):
        """
        Compute equilibration pressures and temperatures for entire suite.
        
        Applies "compute_sample_pressure_temperature" to the "primary"
        property. Result is saved in the "PT" property.
        
        Parameters
        ----------
        method : str
            Code corresponding to desired thermobarometer.
            Current options are:
              - P08: Putirka et al. (2008, Revs. in Min. and Geo.)
              - L09: Lee et al. (2009, EPSL)
              - TGK12_PLG: Till et al. (2012, JGR: Solid Earth), plagioclase
              - TGK12_SPL: Till et al. (2012, JGR: Solid Earth), spinel
              - PF16: Plank and Forsyth (2016, G-cubed)
              - SD20: Sun and Dasgupta (2020, EPSL)
              - BK21: Brown Krien et al. (2021, JGR: Solid Earth), stable phase
              - BK21_PLG: Brown Krien et al. (2021, JGR: Solid Earth), plagioclase
              - BK21_SPL: Brown Krien et al. (2021, JGR: Solid Earth), spinel
              - BK21_GNT: Brown Krien et al. (2021, JGR: Solid Earth), garnet
        min_SiO2 : float
            Threshold SiO2 content below which samples will be ignored.
        """
        self.PT = self.primary.apply(
            compute_sample_pressure_temperature, 
            axis=1, 
            result_type="expand",
            args=(method,min_SiO2,))
    
    def compute_temperature(self, method="PF16", P=1., min_SiO2=0.):
        """
        Compute equilibration temperatures for entire suite at given pressure.
        
        Applies "compute_sample_temperature" to the "primary"
        property. Result is saved in the "PT" property.
        
        Parameters
        ----------
        method : str
            Code corresponding to desired thermometer.
            Current options are:
              - P08: Putirka et al. (2008, Revs. in Min. and Geo.)
              - L09: Lee et al. (2009, EPSL)
              - TGK12_PLG: Till et al. (2012, JGR: Solid Earth), plagioclase
              - TGK12_SPL: Till et al. (2012, JGR: Solid Earth), spinel
              - PF16: Plank and Forsyth (2016, G-cubed)
              - SD20: Sun and Dasgupta (2020, EPSL)
              - BK21: Brown Krien et al. (2021, JGR: Solid Earth), stable phase
              - BK21_PLG: Brown Krien et al. (2021, JGR: Solid Earth), plagioclase
              - BK21_SPL: Brown Krien et al. (2021, JGR: Solid Earth), spinel
              - BK21_GNT: Brown Krien et al. (2021, JGR: Solid Earth), garnet
              - B93: Beattie (1993, Contrib. to Min. and Pet.)
              - P07_2: Putirka et al. (2007, Chem. Geol.), Equation 2
              - P07_4: Putirka et al. (2007, Chem. Geol.), Equation 4
              - HA15: Herzberg and Asimow (2015, G-cubed)
        P : float
            Pressure at which temperature whould be calculated.
        min_SiO2 : float
            Threshold SiO2 content below which samples will be ignored.
        """
        self.PT = self.primary.apply(
            compute_sample_temperature, 
            axis=1, 
            result_type="expand",
            args=(method,P,min_SiO2,)
            )    

    def compute_pressure(self, method="PF16", T=1300., min_SiO2=0.):
        """
        Compute equilibration pressures for entire suite at given temperature.
        
        Applies "compute_sample_pressure" to the "primary"
        property. Result is saved in the "PT" property.
        
        Parameters
        ----------
        method : str
            Code corresponding to desired barometer.
            Current options are:
              - P08: Putirka et al. (2008, Revs. in Min. and Geo.)
              - L09: Lee et al. (2009, EPSL)
              - TGK12_PLG: Till et al. (2012, JGR: Solid Earth), plagioclase
              - TGK12_SPL: Till et al. (2012, JGR: Solid Earth), spinel
              - PF16: Plank and Forsyth (2016, G-cubed)
              - SD20: Sun and Dasgupta (2020, EPSL)
              - BK21: Brown Krien et al. (2021, JGR: Solid Earth), stable phase
              - BK21_PLG: Brown Krien et al. (2021, JGR: Solid Earth), plagioclase
              - BK21_SPL: Brown Krien et al. (2021, JGR: Solid Earth), spinel
              - BK21_GNT: Brown Krien et al. (2021, JGR: Solid Earth), garnet
        T : float
            Temperature at which pressure should be calculated.
        min_SiO2 : float
            Threshold SiO2 content below which samples will be ignored.
        """
        self.PT = self.primary.apply(
            compute_sample_pressure, 
            axis=1, 
            result_type="expand",
            args=(method,T,min_SiO2,)
            )  
        
    def check_samples_for_fitting(self, mantle, filters=(None,), args=((None,))):
        """
        Determine which samples should be fitted.
        
        Samples below the solidus will be dropped. Also applies additional
        filters if provided by the user. Samples which fail any filter will
        be rejected.
        
        Result is saved in "PT_to_fit" property. Same as "PT" but rejected
        samples are assigned nan pressure and temperature.
        
        Parameters
        ----------
        mantle : instance of pyMelt.mantle_class.mantle
            The mantle object to be used to calculate the solidus.
        filters : tuple of functions, optional
            Set of functions to filter samples before fitting.
            Filters take the form of a function that reads a single-row
            dataframe and returns either true or force.
        args : tuple of tuples, optional, same length as filters
            Extra arguments to be passed to the filter functions.
        """
        def combine(df):
            return {'fit': df.to_numpy().all()}

        def above_solidus(df, mantle):
            return {'fit': max([l.TSolidus(df['P']) for l in mantle.lithologies]) - df['T'] < df['T_err']}

        def isnotnan(df):
            return {'fit': ~np.isnan(np.array([df['P'], df['T']])).any()}

        combined = pd.concat([self.data, self.PT], axis=1)
        to_fit = combined.apply(above_solidus, axis=1, result_type="expand", args=(mantle,))
        to_fit = pd.concat([to_fit, combined.apply(isnotnan, axis=1, result_type="expand")], axis=1)
        if filters[0]:
            for f,a in zip(filters,args):
                to_fit = pd.concat([to_fit, combined.apply(f, axis=1, args=a, result_type="expand")], axis=1)
        to_fit = to_fit.apply(combine, axis=1, result_type="expand")
        self.PT['Fit_Tp'] = to_fit.copy()
    
    def find_individual_melt_fractions(self, mantle, path, filters=(None,), filter_args=(None,)):
        """
        Find best-fitting melt fractions for each sample relative to given melt
        path.
        
        Applies "find_sample_melt_fraction" to "PT_to_fit" dataframe. Results
        saved in "individual_melt_fractions" property.
        
        Parameters
        ----------
        mantle : instance of pyMelt.mantle_class.mantle
            The mantle object to be used to calculate the solidus.
        path : instance of pyMelt.meltingcolumn_classes.meltingColumn
            The melting path to be used.
        filters : tuple of functions, optional
            Set of functions to filter samples before fitting.
            Filters take the form of a function that reads a single-row
            dataframe and returns either true or force.
        args : tuple of tuples, optional, same length as filters
            Extra arguments to be passed to the filter functions.
        """
        self.check_samples_for_fitting(mantle, filters, filter_args)
        self.individual_melt_fractions = self.PT.apply(
            find_sample_melt_fraction, 
            axis=1, 
            result_type="expand", 
            args=(path,))

    def find_individual_potential_temperatures(self, mantle, filters=(None,), filter_args=(None,)):
        """
        Find best-fitting melting paths for each sample.
        
        Applies "find_sample_potential_temperature" to "PT_to_fit" dataframe.
        Result is saved in "individual_potential_temperatures" property.

        Parameters
        ----------
        mantle : instance of pyMelt.mantle_class.mantle
            The mantle object to be used to calculate the solidus and melting
            paths.
        filters : tuple of functions, optional
            Set of functions to filter samples before fitting.
            Filters take the form of a function that reads a single-row
            dataframe and returns either true or force.
        args : tuple of tuples, optional, same length as filters
            Extra arguments to be passed to the filter functions.
        """
        self.check_samples_for_fitting(mantle, filters, filter_args)
        self.individual_potential_temperatures = self.PT.apply(
            find_sample_potential_temperature, 
            axis=1, 
            result_type="expand", 
            args=(mantle,))

    def find_suite_potential_temperature(self, mantle, find_bounds=False, bounds_threshold=(2./3.), filters=(None,), filter_args=(None,)):
        """
        Find best-fitting potential temperature for suite.
        
        Uses scipy's minimize_scalar to minimize mean distance from suite
        pressure-temperature estimates to melting path. Options used are:
            method: "bounded"
            bounds:  min --> intersection of solidus with surface
                     max --> max Tp for melting model
            bracket: min --> intersection of solidus with surface
                     max --> max Tp for melting model
        
        Parameters
        ----------
        mantle : instance of pyMelt.mantle_class.mantle
            The mantle object to be used to calculate the solidus and melting
            paths.
        find_bounds : bool, optional
            If True, uses find_bounds to place upper and lower bounds on
            best-fitting potential temperature.
        bounds_threshold : float
            The threshold fraction of points to be lie between the best-fitting
            and bounding temperature melting paths.
        filters : tuple of functions, optional
            Set of functions to filter samples before fitting.
            Filters take the form of a function that reads a single-row
            dataframe and returns either true or force.
        args : tuple of tuples, optional, same length as filters
            Extra arguments to be passed to the filter functions.         
        
        Results (saved as class properties)
        -------
        potential_temperature : float
            The best-fitting potential temperature.
        path : instance of pyMelt.meltingcolumn_classes.meltingColumn
            The best-fitting melting path.
        suite_melt_fractions : pandas dataframe
            Nearest melt fractions, pressures and temperatures along best-
            fitting melting path for each sample.
        upper_potential_temperature : float, if find_bounds is True
            The upper bouding potential temperature.
        lower_potential_temperature : float, if find_bounds is True
            The lower bounding potential temperature.
        upper_path : instance of pyMelt.meltingcolumn_classes.meltingColumn, 
                        if finds_bounds is True
            The upper bounding melting path.
        lower_path : instance of pyMelt.meltingcolumn_classes.meltingColumn, 
                        if finds_bounds is True
            The lower bounding melting path.
        """
        self.check_samples_for_fitting(mantle, filters, filter_args)
        max_Tp = find_max_potential_temperature(mantle)
        Tp_fit = minimize_scalar(
            compute_suite_potential_temperature_misfit, 
            bracket=(min([lith.TSolidus(0.) for lith in mantle.lithologies]),max_Tp), 
            bounds=(min([lith.TSolidus(0.) for lith in mantle.lithologies]),max_Tp),
            args=(self.PT, mantle),
            method="bounded")
        self.path = mantle.adiabaticMelt(Tp_fit.x, Pstart=max(mantle.solidusIntersection(Tp_fit.x))+0.01, dP=-0.01)
        self.suite_melt_fractions = self.PT.apply(find_sample_melt_fraction, axis=1, result_type="expand", args=(self.path,))
        self.potential_temperature = Tp_fit.x
        
        if find_bounds:
            upper_points = []
            lower_points = []
            for i in range(len(self.PT)):
                if self.PT['T'].iloc[i] - self.suite_melt_fractions['T'].iloc[i] > 0:
                    upper_points.append(shp.Point(self.PT['T'].iloc[i], self.PT['P'].iloc[i]))
                elif self.PT['T'].iloc[i] - self.suite_melt_fractions['T'].iloc[i] < 0.:
                    lower_points.append(shp.Point(self.PT['T'].iloc[i], self.PT['P'].iloc[i]))

            self.upper_potential_temperature, self.upper_path = find_bounding_potential_temperature(upper_points, self.potential_temperature, mantle, threshold=bounds_threshold)
            self.lower_potential_temperature, self.lower_path = find_bounding_potential_temperature(lower_points, self.potential_temperature, mantle, lower=True, threshold=bounds_threshold)

    def write_to_csv(self, outfile, write_primary=True, write_PT=True, 
                        write_suite_Tp=False, write_individual_Tp=False):
        """
        Write results to csv.
        
        Combine desired outputs, give appropriate column names, and save as
        csv to a specified location.
        
        Parameters
        ----------
        outfile : str
            Path to location where csv should be saved.
        write_primary : bool
            Whether hydrous wt% concentrations should be saved.
        write_PT : bool
            Whether equilibration pressure/temperature estimates should be
            saved.
        write_suite_Tp : bool
            Whether results of fitting melting paths through suite of pressure/
            temperature estimates should be saved.
        write_individual_Tp : bool
            Whether results of fitting melting paths through individual
            pressure/temperature points should be saved.
        """
        output_df = self.data.copy()
        if write_primary and self.primary is not None:
            output_df = pd.concat([output_df, self.primary], axis=1)
        if write_PT and self.PT is not None:
            output_df = pd.concat([output_df, self.PT], axis=1)
        if write_suite_Tp and self.suite_melt_fractions is not None:
            rename_dict = {
                'F': 'F_suite_fit',
                'P': 'P_suite_fit',
                'T': 'T_suite_fit',
                'misfit': 'misfit_suite_fit'
            }
            suite_out = self.suite_melt_fractions.rename(columns=rename_dict)
            suite_out['Tp_suite_fit'] = self.potential_temperature * self.PT['Fit_Tp']
            suite_out['Tp_max_suite_fit'] = self.upper_potential_temperature * self.PT['Fit_Tp']
            suite_out['Tp_min_suite_fit'] = self.lower_potential_temperature * self.PT['Fit_Tp']
            suite_out = suite_out.replace(0, np.nan)
            output_df = pd.concat([output_df, suite_out], axis=1)
        if write_individual_Tp and self.individual_potential_temperatures is not None:
            rename_dict = {
                'F': 'F_ind_fit',
                'Tp': 'Tp_ind_fit'
            }
            ind_out = self.individual_potential_temperatures.rename(columns=rename_dict)
            ind_out = ind_out.drop(["P", "T", "path", "misfit"], axis=1)
            output_df = pd.concat([output_df, ind_out], axis=1)
        output_df.to_csv(outfile, index=False)


"""
======================
backtrack_compositions
======================

Backtrack sample compositions to 'primary' compositions.
"""

import warnings

import numpy as np
import pandas as pd

import copy

MAJOR_OXIDES = [
    'SiO2','Al2O3','FeO','Fe2O3','MgO','CaO','Na2O','K2O','TiO2','MnO',
    'Cr2O3', 'P2O5', 'NiO', 'CoO', 'H2O', 'CO2']

def normalise(in_comp):
    """
    Normalise dictionary so that values sum to 100%.
    
    Parameters
    ---------
    in_comp : dict
        The values to be normalised.
        
    Returns
    -------
    out_comp : dict
        The normalised values.
    """
    out_comp = {}
    total = sum(in_comp.values())
    for phase in in_comp:
        out_comp[phase] = in_comp[phase] / total * 100.
    return out_comp

def fill_dict_with_nans(in_dict):
    """
    Fill a dictionary with numpy.nan values.
    
    Parameters
    ----------
    in_dict : dict
        The dictionary to be filled with nans.
        
    Returns
    out_dict : dict
        The dictionary with all values replaces with nan.
    """
    out_dict = {}
    for key in in_dict:
        out_dict[key] = np.nan
    return out_dict

class BacktrackOlivineFractionation:
    """
    Correct sample compositions for fractional olivine crystallisation.
    
    Follows the scheme of Lee et al. (2009, EPSL) to correct major-element
    compositions of erupted basalts for the effects of fractional olivine
    crystallisation. Silicon, iron and magnesium in proportions corresponding
    to olvine in equilibrium with the sample are added iteratively until the
    composition reaches some forsterite number representative of the mantle
    source.
    
    The partition coefficient, Kd, can be set to a constant value or be allowed
    to vary as a function of melt magnesium number, according to the scheme of
    Tamura et al. (2000, J. Pet).
    
    Parameters
    ----------
    Kd : float or None
        If float, sets the value of the partition coefficient. If None, the
        partition coefficient is allowed to vary as a function of melt
        magnesium number.
    dm : float
        The mass increment to use during iterative olivine addition.
    verbose : bool
        If True, progress messages are printed during iterative addition of
        olivine.
    max_olivine_addition : float
        The maximum proportion of olivine to add before abandoning.    
    """
    
    def __init__(
        self, Kd=None, dm=0.0005, verbose=False, 
        max_olivine_addition=0.3):
    
        self.fixed_Kd = Kd
        self.dm = dm
        self.verbose = verbose
        self.max_olivine_addition = max_olivine_addition

    @property
    def cation_mole(self):
        """
        Convert oxide concentrations in weight percent to cation concentrations.
            
        Returns
        -------
        out_comp : dict
            The cation concentrations.
        """

        weights = {
            # Combined atomic weights, e.g., SiO2 = Si + O + O = 60.08 (g/mol).
            # Atomic weights given per cation, e.g., Al2O3 is divided by 2.
            'SiO2': 60.08,
            'Al2O3': 101.96 / 2.,
            'FeO': 71.84,
            'Fe2O3': 159.69 / 2.,
            'MgO': 40.3,
            'CaO': 56.08,
            'Na2O': 61.98 / 2.,
            'K2O': 94.2 / 2.,
            'TiO2': 79.86,
            'MnO': 70.94,
            'Cr2O3': 151.99 / 2.,
            'P2O5': 141.942524 / 2.,
            'NiO': 74.69239999999999,
            'CoO': 74.932195,
            'H2O': 18.014680000000002 / 2.,
            'CO2': 44.009
            }
        keys = {
            'SiO2': 'Si',
            'Al2O3': 'Al',
            'FeO': 'FeII',
            'Fe2O3': 'FeIII',
            'MgO': 'Mg',
            'CaO': 'Ca',
            'Na2O': 'Na',
            'K2O': 'K',
            'TiO2': 'Ti',
            'MnO': 'Mn',
            'Cr2O3': 'Cr',
            'P2O5': 'P',
            'NiO': 'Ni',
            'CoO': 'Co',
            'H2O': 'H',
            'CO2': 'C'
        }
        out_comp = {}
        for phase in self.oxide_wt_hydrous:
            out_comp[keys[phase]] = self.oxide_wt_hydrous[phase] / weights[phase]
        return normalise(out_comp)
    
    @property
    def Kd(self):
        """
        Compute the partition coefficient.

        If a fixed value has been specified, it is returned. Otherwise,
        calculate as function of Mg & FeII content, using expression from 
        Tamura et al. (2000, J. Pet), via Lee et al. (2009) spreadsheet.
        """

        if self.fixed_Kd:
            return self.fixed_Kd
        else:
            cation = self.cation_mole
            return 0.25324 + 0.0033663*(cation['Mg'] + 0.33*cation['FeII'])

    @property
    def Fo(self):
        """
        Compute forsterite number from oxide weight compositions.

        Returns
        -------
        Fo : float
            The calculated forsterite number.
        """

        cation = self.cation_mole
        Fo = 1. / (1. + (self.Kd * cation['FeII'] / cation['Mg']))
        return Fo
        
    def add_olivine(self):
        """
        Add olivine in equilibrium with given melt composition.
        
        Returns
        -------
        out_comp : dict
            The updated concentrations.
        """

        Fo = self.Fo
        
        # Compute olivine composition in equilibrium with melt
        oxide_wt_olivine = {
            'FeO': 2. * (1 - Fo) * 71.85,
            'MgO': 2. * Fo * 40.3,
            'SiO2': 60.08
            }
        oxide_wt_olivine = normalise(oxide_wt_olivine)

        # loop over phases adding olivine
        out_comp = {}
        for phase in self.oxide_wt_hydrous:
            if phase == 'SiO2' or phase == 'FeO' or phase == 'MgO':
                out_comp[phase] = (
                    (self.oxide_wt_hydrous[phase] + 
                    self.dm*oxide_wt_olivine[phase]) / 
                    (1. + self.dm)
                    )
            else:
                out_comp[phase] = self.oxide_wt_hydrous[phase] / (1. + self.dm)

        return normalise(out_comp)

    def backtrack_sample_composition(self, df, return_all=False):
        """
        Backtrack composition to desired mantle forsterite number.
    
        Iteratively adds olivine in equilibrium with melt until desired
        composition is reached.
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the initial composition to be backtracked and
            the target forsterite number. Should contain only one row. To use
            with a multi-row dataframe use df.apply().
        return_all : bool
            Return intermediate backtracking compositions.
    
        Returns
        -------
        primary_oxide : df
            The backtracked compositions.
        composition_through_addition : list
            If return_all is True, returns list of dictionaries containing
            intermediate backtracking compositions.
        """

        # Get major oxides from data frame
        self.oxide_wt_hydrous = {}
        for ox in MAJOR_OXIDES:
            self.oxide_wt_hydrous[ox] = df[ox]
        self.oxide_wt_hydrous = normalise(self.oxide_wt_hydrous)

        # Check Fo is below mantle Fo
        if df['src_Fo']-self.Fo < 0.001:
            self.oxide_wt_hydrous = fill_dict_with_nans(self.oxide_wt_hydrous)
            dm_tot = np.nan
            message = df.Sample + ": backtracking failed! Starting Fo above mantle Fo."
            warnings.warn(message)
        # Otherwise add olvine until primary Fo is reached
        else:

            if self.verbose:

                print("Backtracking sample %s to primary composition:" % df.Sample)

            dm_tot = 0.
            composition_through_addition = []
            while df['src_Fo'] - self.Fo > 0.0002:
                
                self.oxide_wt_hydrous = self.add_olivine()
                comp = copy.copy(self.oxide_wt_hydrous)
                comp['Fo'] = self.Fo
                composition_through_addition.append(comp)
                dm_tot += self.dm
                if self.verbose:
                    print(
                        "    - %.2f%% olivine added, melt Fo = %.4f, Kd = %.4f." %
                        (dm_tot/(1.+dm_tot)*100., self.Fo, self.Kd)
                        )
                if dm_tot/(1. + dm_tot) > self.max_olivine_addition:
                    self.oxide_wt_hydrous = fill_dict_with_nans(self.oxide_wt_hydrous)
                    message = (
                        df.Sample + ": backtracking failed! Olivine addition exceeding %d%%" 
                        % (self.max_olivine_addition*100.))
                    warnings.warn(message)
                    break

        # Package up
        primary_oxide = {}
        for phase in self.oxide_wt_hydrous:
            primary_oxide[phase + "_primary_wt"] = self.oxide_wt_hydrous[phase]
            
        # Add amount olivine added
        primary_oxide['ol_added'] = dm_tot / (1. + dm_tot)

        if not return_all:
            return primary_oxide
        else:
            return primary_oxide, composition_through_addition




"""
===============
thermobarometry
===============

Perform thermobarometric calculations.
"""


import warnings
import numpy as np
import pandas as pd
import sympy as sym
import inspect

def Mg_num(df):
    """
    Calculate Mg#.
        
    Parameters
    ----------
    df : pandas dataframe 
       sample compositions in anhydrous mol%.
                
    Returns
    -------
    Mg_num : float
        The Mg#.
    """
    Mg_num = df['MgO_primary_mol_dry'] / (df['MgO_primary_mol_dry'] + 
                                            df['FeO_primary_mol_dry'])
    return Mg_num

def NaK_num(df):
    """
    Calculate NaK#.
    
    Must previously run compute_components_species(),
    compute_components_cation(), compute_components_compound()
    or compute components_per_ox().
    
    Parameters
    ----------
    df : pandas dataframe with sample compositions in anhydrous mole%
            
    Returns
    -------
    NaK_num : float
        The NaK#.        
    """            
    NaK_num = (
        (df['Na2O_primary_wt_dry'] + df['K2O_primary_wt_dry']) / 
        (df['Na2O_primary_wt_dry'] + df['K2O_primary_wt_dry'] + 
            df['CaO_primary_wt_dry'])
        )
    return NaK_num

def normalize_v2(df, list_parameters, input_suffix, output_suffix, end_sum):
    """
    Normalise sample compositions so they sum to a given value.
    
    Concentrations in list_parameters are normalised so that they sum to 
    value given by end_sum. New values added to df with list_parameter
    name appended by output_suffix.
    
    Parameters
    ----------
    df : pandas dataframe
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
    list_parameters : list 
        Dataframe columns to be normalised.
    input_suffix : str
        Original suffix for columns to use.
    output_suffix : str
        Suffix to be added to create new column name.
    end_sum : float
        Value that columns should total after normalisation.

    Returns
    -------
    df : pandas datframe
        The updated dataframe containing normalised values.
    """
    SUM = 0.
    for i in range(0, len(list_parameters), 1):
        parameter = list_parameters[i]
        SUM = df[parameter + input_suffix] + SUM

    for i in range(0, len(list_parameters), 1):
        parameter = list_parameters[i]        
        df[parameter + output_suffix] = end_sum * df[parameter + input_suffix] / SUM
        
    return df

def compute_DMg(df):
    """
    Calculate DMg.

    Follows method outlined in Beattie, (1993), Contrib. Min. and Pet.
    Must previously run compute_components_cation().
    
    Parameters
    ----------
    df : pandas dataframe
        Dataframe containing sample compositions in anhydrous mole%.
            
    Returns
    -------
    DMg : float
        Calculated DMg.
    """ 
    
    Sum_A = (df["FeO_primary_mol_dry"]*0.279)
    Sum_A += (df["MnO_primary_mol_dry"]*0.259)
    Sum_A += (df["MgO_primary_mol_dry"]*1.)
    Sum_A += (df["CaO_primary_mol_dry"]*0.0056)
    Sum_A += (df["NiO_primary_mol_dry"]*3.346)
    #####
    Sum_B = (df["FeO_primary_mol_dry"]*0.031)
    Sum_B -= (df["MnO_primary_mol_dry"]*0.049)
    Sum_B += (df["CaO_primary_mol_dry"]*0.0135)
    Sum_B += (df["NiO_primary_mol_dry"]*-3.665)
    Sum_B += 0.0001*(df["TiO2_primary_mol_dry"] + 
                     df["Al2O3_primary_mol_dry"] +
                     df["Cr2O3_primary_mol_dry"] +
                     df["Fe2O3_primary_mol_dry"] + 
                     df["Na2O_primary_mol_dry"] + 
                     df["K2O_primary_mol_dry"] + 
                     df["P2O5_primary_mol_dry"])
    #####
    DMg = (2./3. - Sum_B) / Sum_A  

    return DMg

def compute_minerals(df):
    """
    Calculate proportions of mineral components.
    
    Using equations listed in Grove (1982, Contrib. Min. Pet., 80:160-182), 
    compute stable mineral assemblage.
        
    Parameters
    ----------
    df : pandas dataframe 
        Datframe containing sample compositions. Expects columns with suffix 
        '_primary_mol_dry'; should be mol% anhydrous compositions.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
                
    Returns
    -------
    MINS: numpy array
        Predicted mineral proportions.
        0: Olivine, 1: cpx, 2: plg, 3: qz, 4: opx, 5: spl, 6: ap.
    """      
    Sum = df["SiO2_primary_mol_dry"] + df["TiO2_primary_mol_dry"] + (df["Al2O3_primary_mol_dry"]*2.) 
    Sum += (df["Cr2O3_primary_mol_dry"]*2) + df["FeO_primary_mol_dry"]
    Sum += df["MnO_primary_mol_dry"] + df["MgO_primary_mol_dry"] 
    Sum += df["CaO_primary_mol_dry"] + (df["Na2O_primary_mol_dry"]*2) + (df["K2O_primary_mol_dry"]*2)
    Sum += df["NiO_primary_mol_dry"] + (df["P2O5_primary_mol_dry"]*2)
    ####    
    sio2 = 100. * df["SiO2_primary_mol_dry"] / Sum
    tio2 = 100. * df["TiO2_primary_mol_dry"] / Sum
    al2o3 = 200. * df["Al2O3_primary_mol_dry"] / Sum
    cr2o3 = 200. * df["Cr2O3_primary_mol_dry"] / Sum
    feo = 100. * df["FeO_primary_mol_dry"] / Sum
    mgo = 100. * df["MgO_primary_mol_dry"] / Sum
    cao = 100. * df["CaO_primary_mol_dry"] / Sum
    na2o = 200. * df["Na2O_primary_mol_dry"] / Sum
    k2o = 200. * df["K2O_primary_mol_dry"] / Sum
    p2o5 = 200. * df["P2O5_primary_mol_dry"] / Sum
    ####     
    Total = sio2 - cao - 2*(k2o+na2o) + cr2o3 + tio2 + (2*p2o5)
    ####    
    ol = (0.5*(feo+mgo+0.5*(al2o3-k2o-na2o)-cao-tio2-cr2o3+5/3*p2o5))/Total*2
    ####
    qz = (sio2-0.5*(feo+mgo)-1.5*cao-0.25*al2o3-2.75*(k2o+na2o)+0.5*cr2o3+0.5*tio2+2.5*p2o5)/Total
    ####  
    cpx = (cao-0.5*al2o3+0.5*(k2o+na2o)-5/3*p2o5)/Total*3
    ####    
    plg = (0.5*(al2o3+na2o-k2o))/Total*4
    ####    
    opx = k2o/Total*4
    #### 
    spl = (cr2o3+tio2)/Total*1.5
    ####   
    ap = p2o5/3/Total*6
    ####     
    Min_Tot = ol + qz + plg + cpx + opx + spl + ap
    ####       
    MINS = [ol/Min_Tot, cpx/Min_Tot, plg/Min_Tot, qz/Min_Tot, opx/Min_Tot, spl/Min_Tot, ap/Min_Tot]

    return MINS

def compute_parameter_BK21(df, OL, P, regression_params):
    """
    Calculate parameter values using equation
    of Brown Krein et al., (2021)
        
    Parameters
    ----------
    df : pandas dataframe
        sample compositions in anhydrous wt% need suffix '_primary_wt_dry'
        sample compositions in anhydrous mol% need suffix '_primary_mol_dry'.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
    OL : flota
        Olivine concentration.
    P: float
        Pressure in GPa.
    regression_params: array
        Array containing eight calibrated parameter values.
                
    Returns
    -------
    result: float
        either T or predicted mineral percentages.
    """ 
    
    result = regression_params[0]
    result += regression_params[1] * OL 
    result += regression_params[2] * P
    result += regression_params[3] * Mg_num(df)
    result += regression_params[4] * NaK_num(df)
    result += regression_params[5] * df['TiO2_primary_wt_dry'] 
    result += regression_params[6] * df['K2O_primary_wt_dry']
    result += regression_params[7] * (df['CaO_primary_wt_dry']/df['Al2O3_primary_wt_dry'])        
    return result

def compute_parameter_TGK12(df, MINS, P, regression_params):
    """
    Calculate parameter values using equation
    of Till et al., (2012)
        
    Parameters
    ----------
    df : pandas dataframe
        sample compositions in anhydrous wt% need suffix '_primary_wt_dry'
        sample compositions in anhydrous mol% need suffix '_primary_mol_dry'.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
    MINS : array
        Concentrations of each mineral.
    P: float
        Pressure in GPa.
    regression_params: array
        one row array of 8 calibrated prameter values.
                
    Returns
    -------
    result: float
        either T or predicted mineral percentages
    """ 
    
    result = regression_params[0]
    result += regression_params[1] * (MINS[0]/(MINS[0]+MINS[1]+MINS[2]+MINS[3]))
    result += regression_params[2] * P
    result += regression_params[3] * (1. - Mg_num(df))
    result += regression_params[4] * NaK_num(df)
    result += regression_params[5] * df['TiO2_primary_wt_dry'] 
    result += regression_params[6] * df['K2O_primary_wt_dry']    
    return result

def compute_components_species(df):
    """
    Calculate mole species proportions.
    
    Use equations listed in Appendix A of Lee et al (2009, G-cubed) to compute
    proportions of each mole species.
        
    Parameters
    ----------
    df : pandas dataframe
        Dataframe containing sample compositions in wt%.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
                
    Returns
    -------
    df : pandas dataframe
        Original dataframe with additional columns:
          - anhydrous wt% with suffix '_primary_wt_dry'
          - oxide mol% with suffix '_primary_mol'
          - species proportions in mol%, e.g., 'Si4O8'
          - anhydrous species proportions in mol%, e.g., 'Si4O8_dry'.
    """   
    
    MAJOR_OXIDES = ['SiO2','Al2O3','FeO','Fe2O3','MgO','CaO','Na2O','K2O',
        'TiO2','MnO','Cr2O3', 'P2O5', 'H2O']
    OXIDES_DRY = ['SiO2','Al2O3','FeO','Fe2O3','MgO','CaO','Na2O','K2O',
        'TiO2','MnO','Cr2O3', 'P2O5']
    OXIDE_WEIGHT = [60.08,101.96,71.84,159.69,40.3,56.08,61.98,
        94.2,79.86,70.94,151.99,141.942524,18.014680000000002]
    SPECIES = ['Si4O8','Al16/3O8','Fe4Si2O8','Fe16/3O8','Mg4Si2O8',
        'Ca4Si2O8','Na2Al2Si2O8','K2Al2Si2O8','Ti4O8','Mn4Si2O8','Cr16/3O8',
        'P16/5O8']     
    SPECIES_DRY = ['Si4O8','Al16/3O8','Fe4Si2O8','Fe16/3O8','Mg4Si2O8',
        'Ca4Si2O8','Na2Al2Si2O8','K2Al2Si2O8','Ti4O8','Mn4Si2O8','Cr16/3O8',
        'P16/5O8']         

    normalize_v2(df, OXIDES_DRY, '_primary_wt', '_primary_wt_dry', 100.)

    for i in range(0, len(MAJOR_OXIDES), 1):
        OXIDE = MAJOR_OXIDES[i]
        df[OXIDE + '_primary_mol'] = df[OXIDE + '_primary_wt'] / OXIDE_WEIGHT[i]

    normalize_v2(df, MAJOR_OXIDES, '_primary_mol', '_primary_mol', 100.)
        
    df['Si4O8'] = (df['SiO2_primary_mol'] - 0.5*(df['FeO_primary_mol']
        +df['MgO_primary_mol']+df['CaO_primary_mol']+df['MnO_primary_mol'])
        - df['Na2O_primary_mol'] - df['K2O_primary_mol']) * 0.25
    df['Al16/3O8'] = (df['Al2O3_primary_mol'] - df['Na2O_primary_mol']) * (3./8.)
    df['Fe4Si2O8'] = df['FeO_primary_mol'] * 0.25
    df['Fe16/3O8'] = df['Fe2O3_primary_mol'] * (3./8.)
    df['Mg4Si2O8'] = df['MgO_primary_mol'] * 0.25
    df['Ca4Si2O8'] = df['CaO_primary_mol'] * 0.25
    df['Na2Al2Si2O8'] = df['Na2O_primary_mol']
    df['K2Al2Si2O8'] = df['K2O_primary_mol']
    df['Ti4O8'] = df['TiO2_primary_mol'] * 0.25
    df['Mn4Si2O8'] = df['MnO_primary_mol'] * 0.25
    df['Cr16/3O8'] = df['Cr2O3_primary_mol']  * (3./8.)
    df['P16/5O8'] = df['P2O5_primary_mol'] * 0.625
    df['H16O8'] = df['H2O_primary_mol'] * 0.125

    normalize_v2(df, SPECIES, '', '', 100.)
    normalize_v2(df, SPECIES_DRY, '', '_dry', 100.)

    return

def compute_components_compound(df):
    """
    Calculate proportions of oxides in mol% assuming all Fe is in FeO
    and the sample is anhydrous.
        
    Parameters
    ----------
    df : pandas dataframe
        The sample compositions in wt%.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
                
    Returns
    -------
    df : pandas dataframe
        Original dataframe with additional columns:
          - anhydrous wt% with suffix '_primary_wt_dry'
          - anhydrous oxide mol% with suffix '_primary_mol_dry'.
    """ 
        
    MAJOR_OXIDES = ['SiO2','Al2O3','FeO','MgO','CaO','Na2O','K2O','TiO2',
        'MnO','Cr2O3','P2O5', 'NiO']
    OXIDE_WEIGHT = [60.08,101.96,71.84,40.3,56.08,61.98,
        94.2,79.86,70.94,151.99,141.945,74.71]

    df['FeO_primary_wt'] = df['FeO_primary_wt'] + 0.8998 * df['Fe2O3_primary_wt']
    df['Fe2O3_primary_wt'] = 0.

    normalize_v2(df, MAJOR_OXIDES, '_primary_wt', '_primary_wt_dry', 100.)

    for i in range(0, len(MAJOR_OXIDES), 1):
        OXIDE = MAJOR_OXIDES[i]
        df[OXIDE + '_primary_mol_dry'] = df[OXIDE + '_primary_wt_dry'] / OXIDE_WEIGHT[i]
    
    normalize_v2(df, MAJOR_OXIDES, '_primary_mol_dry', '_primary_mol_dry', 100.)      
    
    return

def compute_components_cation(df):
    """
    Calculate proportions of cations in mol% assuming the sample is anhydrous.
        
    Parameters
    ----------
    df : pandas dataframe
        Dataframe containing sample compositions in wt%.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
                
    Returns
    -------
    df : pandas dataframe
        Original dataframe with additional columns:
          - anhydrous wt% with suffix '_primary_wt_dry'
          - cation mol% with suffix '_primary_mol'
          - anhydrous cation mol% with suffix '_primary_mol_dry'.
    """
        
    MAJOR_OXIDES = ['SiO2','Al2O3','FeO','Fe2O3','MgO','CaO','Na2O','K2O',
        'TiO2','MnO','Cr2O3', 'P2O5', 'NiO', 'H2O']
    CAT_WEIGHT = [60.08,50.98,71.85,159.96/2.,40.3,56.08,30.99,
        47.1,79.9,70.94,151.99/2.,141.945/2.,74.71, 18.014680000000002]    
    MAJOR_OXIDES_DRY = ['SiO2','Al2O3','FeO','Fe2O3','MgO','CaO','Na2O','K2O',
        'TiO2','MnO','Cr2O3', 'P2O5', 'NiO']
    CAT_WEIGHT_DRY = [60.08,50.98,71.85,159.96/2.,40.3,56.08,30.99,
        47.1,79.9,70.94,151.99/2.,141.945/2.,74.71] 

    normalize_v2(df, MAJOR_OXIDES_DRY, '_primary_wt', '_primary_wt_dry', 100.)    
    normalize_v2(df, MAJOR_OXIDES, '_primary_wt', '_primary_wt', 100.)

    for i in range(0, len(MAJOR_OXIDES_DRY), 1):
        OXIDE = MAJOR_OXIDES_DRY[i]
        df[OXIDE + '_primary_mol_dry'] = df[OXIDE + '_primary_wt_dry'] / CAT_WEIGHT_DRY[i]

    for i in range(0, len(MAJOR_OXIDES), 1):
        OXIDE = MAJOR_OXIDES[i]
        df[OXIDE + '_primary_mol'] = df[OXIDE + '_primary_wt'] / CAT_WEIGHT[i]

    normalize_v2(df, MAJOR_OXIDES, '_primary_mol', '_primary_mol', 1.)     
    normalize_v2(df, MAJOR_OXIDES_DRY, '_primary_mol_dry', '_primary_mol_dry', 1.)  
    
    return

def compute_components_per_oxygen(df):
    """
    Calculate proportions of oxides in mol% assuming all Fe is in FeO
    and the sample is anhydrous.
    Normalise to each to one oxygen per compound.
        
    Parameters
    ----------
    df : pandas dataframe
        Dataframe containing sample compositions in wt%.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().
                
    Returns
    -------
    df : pandas dataframe
        Original dataframe with additional columns:
          - anhydrous wt% with suffix '_primary_wt_dry'
          - volatile free oxide mol content per unit oxygen with suffix '_primary_mol_dry'.
    """ 
    MAJOR_OXIDES = ['SiO2','Al2O3','FeO','MgO','CaO','Na2O','K2O',
        'TiO2','MnO','Cr2O3', 'P2O5', 'NiO', 'H2O', 'CO2'] 
    MAJOR_OXIDES_DRY = ['SiO2','Al2O3','FeO','MgO','CaO','Na2O','K2O','TiO2',
        'MnO','Cr2O3','P2O5', 'NiO']
    OXIDE_WEIGHT = [60.08,101.96,71.84,40.3,56.08,61.98,
        94.2,79.86,70.94,151.99,141.945,74.71]    
    #OXIDE_WEIGHT = [60.843, 101.9612, 71.8444, 40.3039, 56.0774, 61.979,
    #                94.196, 79.8658, 70.9374, 151.9904, 141.945, 74.6928]
    ANO_N = [2., 3., 1., 1., 1., 1., 1., 2., 1., 3., 5., 1.]
    CAT_N = [1., 2., 1., 1., 1., 2., 2., 1., 1., 2., 2., 1.]
    
    df['FeO_primary_wt'] = df['FeO_primary_wt'] + 0.8998 * df['Fe2O3_primary_wt']
    df['Fe2O3_primary_wt'] = 0.

    normalize_v2(df, MAJOR_OXIDES_DRY, '_primary_wt', '_primary_wt_dry', 100.)    
    normalize_v2(df, MAJOR_OXIDES, '_primary_wt', '_primary_wt', 100.)

    SUM = 0.
    for i in range(0, len(MAJOR_OXIDES_DRY), 1):
        OXIDE = MAJOR_OXIDES_DRY[i]
        df[OXIDE + '_primary_mol_dry'] = df[OXIDE + '_primary_wt_dry'] / OXIDE_WEIGHT[i] * ANO_N[i]
        SUM += df[OXIDE + '_primary_mol_dry']
    O_FACTOR = 1. / SUM   

    for i in range(0, len(MAJOR_OXIDES_DRY), 1):
        OXIDE = MAJOR_OXIDES_DRY[i]        
        df[OXIDE + '_primary_mol_dry'] = df[OXIDE + '_primary_mol_dry'] * O_FACTOR / ANO_N[i] * CAT_N[i]
    
    return

class PF16:
    
    def __init__(self, df):
        self.df = df
        self.P_err = 0.24
        self.T_err = 39.

    def compute_water_correction(self):
        """
        Calculate temperature correction from water content.

        Equation (1), Plank and Forsyth (2016, G-cubed).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
            
        Returns
        -------
        correction : pandas dataframe
            The temperature corrections.
        """
        
        correction = (
            40.4*self.df['H2O_primary_wt'] - 
            2.97*self.df['H2O_primary_wt']**2. + 
            0.0761*self.df['H2O_primary_wt']**3.
            )
        return correction

    def compute_CO2_correction(self):
        """
        Calculate temperature correction for samples with equlibration pressures
        above 2 GPa (assumed to be influenced by presence of CO2).

        Equation (3), Plank and Forsyth (2016, G-cubed).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
            
        Returns
        -------
        correction : pandas dataframe
            The temperature corrections.
        """
        correction = (self.df['SiO2_primary_wt_dry'] - 50.3) / -0.12804
        return correction

    def compute_temperature(self):
        """
        Calculate equlibration temperature.

        Equation (1), Plank and Forsyth (2016, G-cubed).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
            
        Returns
        -------
        temperature : pandas dataframe
            The calculated temperature(s).
        """
        temperature = (
            1264.5 +
            7.85*self.df['Mg4Si2O8_dry'] +
            8545./self.df['Si4O8_dry'] -
            5.96*self.df['Al16/3O8_dry']
            )
        return temperature

    def compute_pressure1(self, T):
        """
        Calculate equlibration pressure.

        Equation (2), Plank and Forsyth (2016, G-cubed).

        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        pressure : pandas dataframe
            The calculated pressure(s).
        """
        pressure = (
            np.log(self.df['Si4O8_dry']) -
            4.045 +
            0.0114*self.df['Fe4Si2O8_dry'] +
            0.00052*self.df['Ca4Si2O8_dry']**2. +
            0.0024*self.df['Mg4Si2O8_dry']) / (-336.3/T - 0.0007*np.sqrt(T)
            )
        return pressure

    def compute_pressure(self, T):
        """
        Calculate equlibration pressure.

        Equation (2), Plank and Forsyth (2016, G-cubed).

        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        T : float
            Temperature in oC.
            
        Returns
        -------
        pressure : pandas dataframe
            The calculated pressure(s).
        """
        compute_components_species(self.df)
        
        P = (
            np.log(self.df['Si4O8_dry']) -
            4.045 +
            0.0114*self.df['Fe4Si2O8_dry'] +
            0.00052*self.df['Ca4Si2O8_dry']**2. +
            0.0024*self.df['Mg4Si2O8_dry']) / (-336.3/(T+273.15) - 0.0007*np.sqrt(T+273.15)
            )
                                               
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': T * self.P_err/P}

    def compute_pressure_temperature(self):
        """
        Compute equilibration pressure and temperature for a given sample.

        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """
        compute_components_species(self.df)
        water_correction = self.compute_water_correction()
        T = self.compute_temperature() - water_correction
        P = self.compute_pressure1(T)
        if P > 2.:
            T -= self.compute_CO2_correction()
            P = self.compute_pressure1(T)
        out = {'P': P, 'T': T - 273.15, 'P_err': self.P_err, 'T_err': self.T_err}
        return out

class L09:
        
    def __init__(self, df):
        self.df = df
        self.P_err = 0.2
        self.T_err = 40.
        
    def compute_pressure1(self, T):
        """
        Calculate equlibration temperature.

        Equation (2), Lee et al (2009, G-cubed).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions in species mol%.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        T : float
            temperature in oC
            
        Returns
        -------
        pressure : pandas dataframe
            The calculated pressure(s).
        """
        
        pressure = (
            (np.log(self.df['Si4O8']) - 
                4.019 + 
                0.0165*self.df['Fe4Si2O8'] + 
                0.0005*(self.df['Ca4Si2O8']**2.)) /
            (-770*(T**(-1.)) + 0.0058*(T**0.5) - 0.003*self.df['H16O8'])
            )
        return pressure
    
    def compute_pressure(self, T):
        """
        Calculate equlibration temperature.

        Equation (2), Lee et al (2009, G-cubed).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions in species mol%.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        T : float
            temperature in oC
            
        Returns
        -------
        pressure : pandas dataframe
            The calculated pressure(s).
        """
        compute_components_species(self.df)
        P = (
            (np.log(self.df['Si4O8']) - 
                4.019 + 
                0.0165*self.df['Fe4Si2O8'] + 
                0.0005*(self.df['Ca4Si2O8']**2.)) /
            (-770*((T+273.15)**(-1.)) + 0.0058*(T**0.5) - 0.003*self.df['H16O8'])
            )  
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': T * self.P_err/P}
    
    def compute_temperature(self):
        """
        Calculate equlibration temperature.

        Equation (3), Lee et al (2009, G-cubed).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        temperature : pandas dataframe
            The calculated temperature(s).
        """
        
        temperature = (
            916.45 + 
            (13.68 * self.df['Mg4Si2O8']) + 
            (4580. / self.df['Si4O8']) -
            (0.509 * self.df['H16O8'] * self.df['Mg4Si2O8'])
            )
        return temperature
        
    def compute_pressure_temperature(self):
        """
        Compute equilibration pressure and temperature for a given sample.

        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """
        
        compute_components_species(self.df)
        T = self.compute_temperature()
        P = self.compute_pressure1((T+273.15))
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err}

class TGK12_SPL:
    
    def __init__(self, df):
        self.df = df
        self.P_err = 0.15
        self.T_err = 10.78

    def compute_temperature(self, P):
        """
        Compute equilibration pressure and temperature for a given sample.
        
        Assumes spinel is stable mantle phase.
        Lines (3) and (4) of Table (5), Till et al (2012, JGR).
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : 
            Pressure in GPa.
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """
        
        compute_components_compound(self.df)       
        MINS = compute_minerals(self.df)
        T = compute_parameter_TGK12(self.df, MINS, P, [1212,0.,119.9,-97.33,-87.76,3.44,-4.58])  
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err}         
        

    def compute_pressure_temperature(self):    
        """
        Compute equilibration pressure and temperature for a given sample.
        
        Assumes spinel is stable mantle phase.
        Lines (3) and (4) of Table (5), Till et al (2012, JGR).
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """
        
        compute_components_compound(self.df)       
        MINS = compute_minerals(self.df)
        P = compute_parameter_TGK12(self.df, MINS, 0., [-0.862,9.471,0.,-2.383,2.922,0.218,-0.146])
        T = compute_parameter_TGK12(self.df, MINS, P, [1212,0.,119.9,-97.33,-87.76,3.44,-4.58])  
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err}   
    
class TGK12_PLG:
 
    def __init__(self, df):
        self.df = df
        self.P_err = 0.08
        self.T_err = 11.7
    
    def compute_temperature(self, P):
        """
        Compute equilibration pressure and temperature for a given sample.
        
        Assumes plagioclase is stable mantle phase.
        Line (2) of Table (5), Till et al (2012, JGR).
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
            Pressure in GPa.
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """
        
        compute_components_compound(self.df)
        MINS = compute_minerals(self.df)
        T = compute_parameter_TGK12(self.df, MINS, P, [1216,0.,104.4,-72.83,-194.9,24.08,-1.55])        
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err}        

    def compute_pressure_temperature(self):
        """
        Compute equilibration pressure and temperature for a given sample.
        
        Assumes plagioclase is stable mantle phase.
        Line (1 and 2) of Table (5), Till et al (2012, JGR).
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """
        
        compute_components_compound(self.df)
        MINS = compute_minerals(self.df)
        P = compute_parameter_TGK12(self.df, MINS, 0., [-1.64,12.94,0.,-2.363,3.51,0.152,-0.176])
        T = compute_parameter_TGK12(self.df, MINS, P, [1216,0.,104.4,-72.83,-194.9,24.08,-1.55])        
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err}  

class G13:

    def __init__(self, df):
        self.df = df
        self.P_err = 0.24
        self.T_err = 24.

    def compute_pressure(self):
        """
        Calculate equlibration pressure.

        Line (1) of Table (4), Grove et al (2013, Contrib Min Pet).

        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
            
        Returns
        -------
        pressure : float
            The calculated pressure(s).
        """
        compute_components_compound(self.df) 
        MINS = compute_minerals(self.df)
        pressure = -17.3
        pressure += 146.4 * MINS[0]
        pressure -= 38.4 * (1. - Mg_num(self.df)) 
        pressure += 19.6 * NaK_num(self.df)
        pressure += 4.81 * self.df['P2O5_primary_wt_dry'] 
        pressure *= 0.1
        return pressure

    def compute_temperature1(self, P):
        """
        Calculate equlibration temperature.

        Line (6) of Table (4), Grove et al (2013, Contrib Min Pet).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
        P: pressure in GPa
            
        Returns
        -------
        temperature : float
            The calculated temperature(s) in oC.
        """
        temperature = 1313.67
        temperature += 8.423 * ((P*10.)-1.)
        temperature -= 149.92 * (1. - Mg_num(self.df)) 
        temperature += 55.02 * NaK_num(self.df)
        temperature -= 59.69 * self.df['P2O5_primary_wt_dry'] 
        return temperature

    def compute_temperature(self, P):
        """
        Calculate equlibration temperature.

        Line (6) of Table (4), Grove et al (2013, Contrib Min Pet).
        
        Parameters
        ----------
        df : pandas dataframe
            The sample compositions.
        P: pressure in GPa
            
        Returns
        -------
        temperature : float
            The calculated temperature(s) in oC.
        """
        compute_components_compound(self.df)  
        T = 1313.67
        T += 8.423 * ((P/0.1)-1)
        T -= 149.92 * (1. - Mg_num(self.df))
        T += 55.02 * NaK_num(self.df)
        T -= 59.69 * self.df['P2O5_primary_wt_dry']
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 
    
    def compute_pressure_temperature(self):
        """
        Compute equilibration pressure and temperature for a given sample.
        
        Assumes garnet is stable mantle phase.
        Line (1) and (6) of Table (4), Grove et al (2013, Contrib Min Pet).
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """      
        compute_components_compound(self.df)
        P = self.compute_pressure()
        T = self.compute_temperature1(P)
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err} 

class BK21_GNT:

    def __init__(self, df):
        self.df = df
        self.P_err = 1.26
        self.T_err = 18.72

    def compute_temperature(self, P):
        """
        Calculate equlibration temperature.
        
        Equation 1G-T of Table (1), Brown Krein et al (2021, JGR: Solid Earth).
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
        Pressure in GPa
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """                  
        T = compute_parameter_BK21(self.df, 0., P/0.1, [1136.36052,0,8.73915,184.88412,-19.48245,29.00187,-23.41730,-22.48205])
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 
    
    def compute_pressure_temperature(self):
        """
        Same as TGK12_GNT but equations have been recalibrated with additional data.

        Equations 1G-P and 1G-T of Table (1), Brown Krein et al (2021, JGR: Solid Earth).
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """       
        compute_components_compound(self.df)
        MINS = compute_minerals(self.df)
        P = 0.1 * compute_parameter_BK21(self.df, MINS[0], 0., [-77.53423,139.87872,0,46.43963,28.72057,1.52179,3.25095,18.62211])
        T = compute_parameter_BK21(self.df, 0., P/0.1, [1136.36052,0,8.73915,184.88412,-19.48245,29.00187,-23.41730,-22.48205])
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err} 
    
class BK21_SPL:

    def __init__(self, df):
        self.df = df
        self.P_err = 1.53
        self.T_err = 14.28

    def compute_temperature(self, P):
        """
        Calculate equlibration temperature.
        
        Line (17) of Table (1), Brown Krein et al (2021, JGR: Solid Earth).
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
        Pressure in GPa
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """                  
        T = compute_parameter_BK21(self.df, 0., P/0.1, [1049.24929,0,12.71243,63.47484,-3.32516,2.65802,-12.03128,117.75713])
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 

    def compute_pressure_temperature(self):
        """
        Same as TGK12_SPL but equations have been recalibrated with additional data.

        Line (3 and 17) of Table (1), Brown Krein et al (2021, JGR: Solid Earth).
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """       
        compute_components_compound(self.df)
        MINS = compute_minerals(self.df)
        P = 0.1 * compute_parameter_BK21(self.df, MINS[0], 0., [-29.5,84.82,0,25.,24.67,2.79,-0.138,-1.848]) 
        T = compute_parameter_BK21(self.df, 0., P/0.1, [1049.24929,0,12.71243,63.47484,-3.32516,2.65802,-12.03128,117.75713])
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err} 

class BK21_PLG:
   
    def __init__(self, df):
        self.df = df
        self.P_err = 0.79
        self.T_err = 11.5
        
    def compute_temperature(self, P):
        """
        Calculate equlibration temperature.
        
        Line (6) of Table (1), Brown Krein et al (2021, JGR: Solid Earth).
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
        Pressure in GPa
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """                  
        T = compute_parameter_BK21(self.df, 0., P/0.1, [1074.38633,0,11.86431,65.55420,-138.22714,20.55173,5.85532,79.01883])
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 
    
    def compute_pressure_temperature(self):
        """
        Calculate equlibration conditions.
        
        Line (1 and 7) of Table (1), Brown Krein et al (2021, JGR: Solid Earth).
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """     
        compute_components_compound(self.df)
        MINS = compute_minerals(self.df)
        P = 0.1 * compute_parameter_BK21(self.df, MINS[0], 0., [-43.58532,136.94746,0,24.54202,37.54688,2.10967,-0.77250,1.63584])
        T = compute_parameter_BK21(self.df, 0., P/0.1, [1074.38633,0,11.86431,65.55420,-138.22714,20.55173,5.85532,79.01883])
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err} 

class BK21:
    
    def __init__(self, df):
        self.df = df
        self.P_err = [0.79, 1.53, 1.26]
        self.T_err = [11.5, 14.28, 18.72]
    
    def compute_pressure(self):
        """
        Calculate equlibration pressure.

        Brown Krein et al (2021, JGR: Solid Earth).

        Parameters
        ----------
        df : pandas dataframe
            The sample compositions in wt%.
            
        Returns
        -------
        pressure : array
            The calculated pressure(s) in GPa
            in plg, spl and gnt stability fields.
        """
        MINS = compute_minerals(self.df)
        pressure = [0]*3
        # Compute Pressure Plag        
        pressure[0] = 0.1 * compute_parameter_BK21(self.df, MINS[0], 0., [-43.58532,136.94746,0,24.54202,37.54688,2.10967,-0.77250,1.63584]) 
        # Compute Pressure Spl
        pressure[1] = 0.1 * compute_parameter_BK21(self.df, MINS[0], 0., [-29.5,84.82,0,25.,24.67,2.79,-0.138,-1.848]) 
        # Compute Pressure Gnt 
        pressure[2] = 0.1 * compute_parameter_BK21(self.df, MINS[0], 0., [-77.53423,139.87872,0,46.43963,28.72057,1.52179,3.25095,18.62211])
  
        return pressure

    def compute_temperature1(self, P):
        """
        Calculate equlibration temperature.

        Brown Krein et al (2021, JGR: Solid Earth).

        Parameters
        ----------
        df : pandas dataframe
            The sample compositions in wt%.
        P : float
            Pressure in GPa
                
        Returns
        -------
        temperature : array
            The calculated pressure(s) in GPa 
            in plg, spl and gnt stability fields.
        """
        temperature = [0]*3
        # Compute Temperature Plag
        temperature[0] = compute_parameter_BK21(self.df, 0., P[0]/0.1, [1074.38633,0,11.86431,65.55420,-138.22714,20.55173,5.85532,79.01883])
        # Compute Temperature Spl
        temperature[1] = compute_parameter_BK21(self.df, 0., P[1]/0.1, [1049.24929,0,12.71243,63.47484,-3.32516,2.65802,-12.03128,117.75713])           
        # Compute Temperature Gnt
        temperature[2] = compute_parameter_BK21(self.df, 0., P[2]/0.1, [1136.36052,0,8.73915,184.88412,-19.48245,29.00187,-23.41730,-22.48205])           
                  
        return temperature
    
    def compute_temperature(self, P):
        """
        Calculate equlibration temperature assuming a known pressure.

        Brown Krein et al (2021, JGR: Solid Earth).

        Parameters
        ----------
        df : pandas dataframe
            The sample compositions in wt%.
        P : float
            Pressure in GPa
            
        Returns
        -------
        temperature : array
            The calculated pressure(s) in GPa 
            in plg, spl and gnt stability fields.
        """
        compute_components_compound(self.df)
        P = self.compute_pressure()
        RMSD = self.compute_mins(P)
        T = [0]*3
        # Compute Temperature Plag
        T[0] = compute_parameter_BK21(self.df, 0., P[0]/0.1, [1074.38633,0,11.86431,65.55420,-138.22714,20.55173,5.85532,79.01883])
        # Compute Temperature Spl
        T[1] = compute_parameter_BK21(self.df, 0., P[1]/0.1, [1049.24929,0,12.71243,63.47484,-3.32516,2.65802,-12.03128,117.75713])           
        # Compute Temperature Gnt
        T[2] = compute_parameter_BK21(self.df, 0., P[2]/0.1, [1136.36052,0,8.73915,184.88412,-19.48245,29.00187,-23.41730,-22.48205])           
                  
        return {'P': P[np.argmin(RMSD)], 'T': T[np.argmin(RMSD)],
                'P_err': P[np.argmin(RMSD)] * self.T_err[np.argmin(RMSD)]/T[np.argmin(RMSD)], 
                'T_err': self.T_err[np.argmin(RMSD)]}   

    def compute_mins(self, P):
        """
        Determine root mean squared differences between predicted
        sample mineralogy and mineralogies in plg, spl and gnt fields.

        Brown Krein et al (2021, JGR: Solid Earth).

        Parameters
        ----------
        df : pandas dataframe
            The sample compositions in wt%.
        P: float
            pressure in GPa
        
        Returns
        -------
        RMSD : array
            Difference between calculated and predicted
            mineralogic make-up for plg, spl and gnt fields.
        """
        MINS = compute_minerals(self.df)
        RMSD = [0]*3
        # Calculated minerals order: OL, CPX, PLG, QTZ
        # Mineral Residuals for Plagioclase
        Mplg = [0]*4
        Mplg[0] = compute_parameter_BK21(self.df, 0., P[0]/0.1, [0.412,0,0.005,-0.162,-0.363,-0.011,-0.001,-0.126])
        Mplg[0] -= MINS[0]
        Mplg[1] = compute_parameter_BK21(self.df, 0., P[0]/0.1, [-0.256,0,0,0.032,0.339,-0.004,-0.01,0.497])
        Mplg[1] -= MINS[1]        
        Mplg[2] = compute_parameter_BK21(self.df, 0., P[0]/0.1, [0.429,0,0.012,0.132,0.503,0.001,-0.073,-0.157])        
        Mplg[2] -= MINS[2]        
        Mplg[3] = compute_parameter_BK21(self.df, 0., P[0]/0.1, [0.423,0,-0.018,0.001,-0.487,-0.001,0.023,-0.225])       
        Mplg[3] -= MINS[3]
        
        RMSD[0] = np.sqrt((Mplg[0]**2. + Mplg[1]**2. + Mplg[2]**2. + Mplg[3]**2.)/4.)
        
        # Mineral Residuals for Spinel
        Mspl = [0]*4
        Mspl[0] = compute_parameter_BK21(self.df, 0., P[1]/0.1, [0.294,0,0.009,-0.215,-0.208,-0.027,-0.003,-0.046])
        Mspl[0] -= MINS[0]        
        Mspl[1] = compute_parameter_BK21(self.df, 0., P[1]/0.1, [-0.256,0,0,0.016,0.336,-0.001,-0.006,0.525])
        Mspl[1] -= MINS[1]        
        Mspl[2] = compute_parameter_BK21(self.df, 0., P[1]/0.1, [0.826,0,-0.004,0.116,0.392,-0.012,-0.072,-0.482]) 
        Mspl[2] -= MINS[2]        
        Mspl[3] = compute_parameter_BK21(self.df, 0., P[1]/0.1, [0.148,0,-0.005,0.085,-0.53,0.023,0.02,-0.107])  
        Mspl[3] -= MINS[3]    
        
        RMSD[1] = np.sqrt((Mspl[0]**2. + Mspl[1]**2. + Mspl[2]**2. + Mspl[3]**2.)/4.)
         
        # Mineral Residuals for Garnet
        Mgnt = [0]*4
        Mgnt[0] = compute_parameter_BK21(self.df, 0., P[2]/0.1, [0.531,0,0.007,-0.314,-0.185,-0.01,-0.023,-0.107])
        Mgnt[0] -= MINS[0]          
        Mgnt[1] = compute_parameter_BK21(self.df, 0., P[2]/0.1, [-0.192,0,-0.003,0.109,0.358,-0.005,-0.011,0.413])
        Mgnt[1] -= MINS[1]          
        Mgnt[2] = compute_parameter_BK21(self.df, 0., P[2]/0.1, [0.613,0,-0.006,0.237,0.365,-0.018,-0.042,-0.243]) 
        Mgnt[2] -= MINS[2]          
        Mgnt[3] = compute_parameter_BK21(self.df, 0., P[2]/0.1, [0.048,0,0.003,-0.026,-0.547,0.019,0.010,-0.069])  
        Mgnt[3] -= MINS[3]   

        RMSD[2] = np.sqrt((Mgnt[0]**2. + Mgnt[1]**2. + Mgnt[2]**2. + Mgnt[3]**2.)/4.)
        
        return RMSD      
    
    def compute_pressure_temperature(self):
        """
        Calculate equlibration conditions.
        
        Brown Krein et al (2021, JGR: Solid Earth).
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """    
        compute_components_compound(self.df)
        P = self.compute_pressure()
        T = self.compute_temperature1(P)
        RMSD = self.compute_mins(P)
        return {'P': P[np.argmin(RMSD)], 'T': T[np.argmin(RMSD)],
                'P_err': self.P_err[np.argmin(RMSD)], 'T_err': self.T_err[np.argmin(RMSD)]} 

class HA15:   

    def __init__(self, df):
        self.df = df
        self.P_err = np.nan
        self.T_err = 31.
    
    def compute_temperature(self, P):
        """
        Calculate temperate of olivine liquidus at 1 bar
        using Beattie (1993) Cointrib Min & Pet 115:103-111 
        Equations 10 and 12.

        Convert to desired pressure using Equation from 
        Herzberg & O'Hara, (2002) G3 9:9  
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
            assumed pressure in GPa.
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """    
        compute_components_cation(self.df)
        DMg = compute_DMg(self.df)   
        
        #####
        Temp_Denom = 52.05/8.3143
        Temp_Denom += 2.*np.log(DMg)
        Temp_Denom += 2.*np.log(1.5*(self.df["FeO_primary_mol_dry"]+self.df["MnO_primary_mol_dry"]+self.df["MgO_primary_mol_dry"]+self.df["CaO_primary_mol_dry"]+self.df["NiO_primary_mol_dry"]))
        Temp_Denom += 2.*np.log(3.*self.df["SiO2_primary_mol_dry"])
        Temp_Denom -= 3.5*np.log(1-self.df["Al2O3_primary_mol_dry"]) + 7*np.log(1-self.df["TiO2_primary_mol_dry"])
        Temp_1bar = (113100/8.3143 + P*0.00000411/8.3143)
        Temp_1bar /= Temp_Denom
        Temp_1bar -= 273.15
        #####
        T = Temp_1bar + 54.*P - 2.*P**2.
        
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 

class P08:
    
    def __init__(self, df):
        self.df = df
        self.P_err = 0.29
        self.T_err = 52.
    
    def compute_pressure(self, T):
        """
        Calculate temperate and pressure of melt
        equilibration by simultaneously solving
        Equation 4 of Putirka et al., (2007, Chemical Geology)
        and Equation 42 of Putirka (2008).
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        T : float
            Temperature in oC
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """    
        
        compute_components_cation(self.df)
        aSiO2 = (3*self.df["SiO2_primary_mol_dry"])**-2*(1-self.df["Al2O3_primary_mol_dry"])**(7/2)*(1-self.df["TiO2_primary_mol_dry"])**(7/2)
        P = (231.5+0.186*T+0.1244*T*(np.log(aSiO2))-528.*(aSiO2)**0.5+103.3*self.df["TiO2_primary_mol_dry"]+69.9*(self.df["Na2O_primary_mol_dry"]+self.df["K2O_primary_mol_dry"])+77.3*(self.df["Al2O3_primary_mol_dry"]/(self.df["Al2O3_primary_mol_dry"]+self.df["SiO2_primary_mol_dry"])))/10.
            
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': T * self.P_err/P} 

    
    def compute_pressure_temperature(self):
        """
        Calculate temperate and pressure of melt
        equilibration by simultaneously solving
        Equation 4 of Putirka et al., (2007, Chemical Geology)
        and Equation 42 of Putirka (2008).
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """    
        
        compute_components_cation(self.df)
        DMg = compute_DMg(self.df)   
        aSiO2 = (3*self.df["SiO2_primary_mol_dry"])**-2*(1-self.df["Al2O3_primary_mol_dry"])**(7/2)*(1-self.df["TiO2_primary_mol_dry"])**(7/2)

        P, T = sym.symbols('P, T')
        eq1 = sym.Eq(1 / ((np.log(DMg) + 2.158 - 5.115*10**-2*(self.df["Na2O_primary_wt"]+self.df["K2O_primary_wt"]) + 6.213*10**-2*self.df["H2O_primary_wt"]) / (55.09*P + 4430.)), T)
        eq2 = sym.Eq((231.5+0.186*T+0.1244*T*(np.log(aSiO2))-528.*(aSiO2)**0.5+103.3*self.df["TiO2_primary_mol_dry"]+69.9*(self.df["Na2O_primary_mol_dry"]+self.df["K2O_primary_mol_dry"])+77.3*(self.df["Al2O3_primary_mol_dry"]/(self.df["Al2O3_primary_mol_dry"]+self.df["SiO2_primary_mol_dry"])))/10., P)
        result = sym.solve([eq1,eq2],(P,T))
        T = result[T]
        P = result[P]
            
        return {'P': P, 'T': T, 'P_err': self.P_err, 'T_err': self.T_err} 

class SD20:

    def __init__(self, df):
        self.df = df
        self.P_err = 0.5
        self.T_err = 49.   
    
    def compute_temperature(self, P):
        """
        Sun & Dasgupta, (2020), EPSL.
        Temperature calculated using equation 6.
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
            pressure in GPa
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """          

        compute_components_per_oxygen(self.df)
        
        # Phi and Theta defined in Eq 3,
        # Omega defined in Eq 5
        phi = 2412. + 9.2 * (self.df['CO2_primary_wt']/self.df['SiO2_primary_mol_dry'])
        phi += 11.19 * (self.df['H2O_primary_wt']/self.df['SiO2_primary_mol_dry'])
        
        theta = 4.7 + 3.91 * np.log(self.df['SiO2_primary_mol_dry']) 
        theta += 1.19*((self.df['K2O_primary_mol_dry']+self.df['TiO2_primary_mol_dry'])/self.df['MgO_primary_mol_dry'])
        
        omega = 2.59 + 3.5*(self.df['CaO_primary_mol_dry'] - (2.*self.df['K2O_primary_mol_dry'])) 
        omega += 4.85*self.df['TiO2_primary_mol_dry'] 
        omega += 1.4*(self.df['MgO_primary_mol_dry']/(self.df['MgO_primary_mol_dry']+self.df['FeO_primary_mol_dry'])) 
        omega += 0.5 * self.df['MgO_primary_mol_dry'] * sym.sqrt(self.df['CO2_primary_wt']) 
        omega += 0.057 * self.df['H2O_primary_wt']      
        
        T = (10.**4. / (omega - 0.34*np.sqrt(P) - 1.26*np.log(self.df['MgO_primary_mol_dry']))) - 273.15
        
        return {'P': P, 'T': float(T), 'P_err': float(P * self.T_err/T), 'T_err': self.T_err} 
    
    def compute_pressure_temperature(self):
        """
        Sun & Dasgupta, (2020), EPSL.
        Pressure and temperature calculated using
        equations 4 and 6, repsectively.
        Pressure and temperature are functions of each other
        and must be calculated simultaneously.
    
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
            
        Returns
        -------
        out : dict
            The equilibration pressure and temperature result.
            Temperature is in degrees celcius.
            Pressure is in GPa.
            Associated errors for pressure and temperature.
        """          

        compute_components_per_oxygen(self.df)
        
        # Phi and Theta defined in Eq 3,
        # Omega defined in Eq 5
        phi = 2412. + 9.2 * (self.df['CO2_primary_wt']/self.df['SiO2_primary_mol_dry'])
        phi += 11.19 * (self.df['H2O_primary_wt']/self.df['SiO2_primary_mol_dry'])
        
        theta = 4.7 + 3.91 * np.log(self.df['SiO2_primary_mol_dry']) 
        theta += 1.19*((self.df['K2O_primary_mol_dry']+self.df['TiO2_primary_mol_dry'])/self.df['MgO_primary_mol_dry'])
        
        omega = 2.59 + 3.5*(self.df['CaO_primary_mol_dry'] - (2.*self.df['K2O_primary_mol_dry'])) 
        omega += 4.85*self.df['TiO2_primary_mol_dry'] 
        omega += 1.4*(self.df['MgO_primary_mol_dry']/(self.df['MgO_primary_mol_dry']+self.df['FeO_primary_mol_dry'])) 
        omega += 0.5 * self.df['MgO_primary_mol_dry'] * sym.sqrt(self.df['CO2_primary_wt']) 
        omega += 0.057 * self.df['H2O_primary_wt']        

        # Pressure calculated using Eq 4.
        # Temperature calculated using Eq 6.
        # Must be solved simultaneously.       
        P, T = sym.symbols('P, T')
        eq1 = sym.Eq(13. - sym.sqrt((T/45.9) * (np.log(self.df['Al2O3_primary_mol_dry']/self.df['MgO_primary_mol_dry']) - theta) - (phi/45.9) + 169.), P)
        eq2 = sym.Eq(10.**4. / (omega - 0.34*sym.sqrt(P) - 1.26*np.log(self.df['MgO_primary_mol_dry'])), T)
        result = sym.solve([eq1,eq2],(P,T))
        if len(result) == 0:
            T = np.nan
            P = np.nan
        else:
            T = result[0][1] - 273.15
            P = result[0][0]
        return {'P': float(P), 'T': float(T), 'P_err': self.P_err, 'T_err': self.T_err} 

class B93:

    def __init__(self, df):
        self.df = df
        self.T_err = 20.
    
    def compute_temperature(self, P):
        """
        Calculate temperate of olivine liquidus at P GPa
        using Equation 10 Beattie (1993, Contrib. to Min. and Pet.)
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
            pressure in GPa
            
        Returns
        -------
        out : dict
            The equilibration temperature result.
            Temperature is in degrees celcius.
            Assumed pressure is in GPa.
            Associated errors for pressure and temperature.
        """    
        compute_components_cation(self.df)
        DMg = compute_DMg(self.df)       
        
        # Eq 20 of Putirka 2008 Rev in Min. & Geochem. 69:1
        # DMg = (0.666-(-0.049*self.df["MnO_primary_mol"]+0.027*self.df["FeO_primary_mol"]))/(self.df["MgO_primary_mol"]+0.259*self.df["MnO_primary_mol_dry"]+0.299*self.df["FeO_primary_mol"])
        C_NM = self.df["MgO_primary_mol"] + self.df["FeO_primary_mol"] 
        C_NM += self.df["CaO_primary_mol"] + self.df["MnO_primary_mol"]
        NF = 7/2*np.log(1.-self.df["Al2O3_primary_mol"]) 
        NF += 7.*np.log(1.-self.df["TiO2_primary_mol"])

        # Eq 19 of Putirka 2008 Rev in Min. & Geochem. 69:1
        T = ((113100./8.3144) + ((P*10.**9.-10.**-5.)*((4.11*10**-6)/8.3144)))
        T /= ((52.05/8.3144) + 2.*np.log(DMg) + 2.*np.log(1.5*C_NM) + 2.*np.log(3.*self.df["SiO2_primary_mol"]) - NF)
        T -= 273.15
  
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 

class P07_2:

    def __init__(self, df):
        self.df = df
        self.T_err = 52.
    
    def compute_temperature(self, P):
        """
        Calculate temperate of olivine liquidus at P GPa
        using Equation 2 of Putirka et al., (2007, Chemical Geology)
        
        Parameters
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
            pressure in GPa
            
        Returns
        -------
        out : dict
            The equilibration temperature result.
            Temperature is in degrees celcius.
            Assumed pressure is in GPa.
            Associated errors for pressure and temperature.
        """
        compute_components_cation(self.df)
        DMg = compute_DMg(self.df)   

        # Eq 21 of Putirka 2008 Rev in Min. & Geochem. 69:1
        T = 1 / ((np.log(DMg) + 2.158 - 5.115*10**-2*(self.df["Na2O_primary_wt"]+self.df["K2O_primary_wt"]) + 6.213*10**-2*self.df["H2O_primary_wt"]) / (55.09*P + 4430))
  
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 

class P07_4:

    def __init__(self, df):
        self.df = df
        self.T_err = 52.
    
    def compute_temperature(self, P):
        """
        Calculate temperate of olivine liquidus at P GPa
        using Equation 4 of Putirka et al., (2007, Chemical Geology)
        
        Parameters
        ----------
        df : pandas dataframe
            Dataframe containing the sample primary composition.
            Should contain only one row. To use with a multi-row dataframe use
            df.apply().
        P : float
            pressure in GPa
            
        Returns
        -------
        out : dict
            The equilibration temperature result.
            Temperature is in degrees celcius.
            Assumed pressure is in GPa.
            Associated errors for pressure and temperature.    
        """
        
        compute_components_cation(self.df)
        DMg = compute_DMg(self.df)   

        C_NM = self.df["MgO_primary_mol"] + self.df["FeO_primary_mol"] 
        C_NM += self.df["CaO_primary_mol"] + self.df["MnO_primary_mol"]
        
        NF = 7/2*np.log(1.-self.df["Al2O3_primary_mol"]) 
        NF += 7.*np.log(1.-self.df["TiO2_primary_mol"])

        # Eq 22 of Putirka 2008 Rev in Min. & Geochem. 69:1        
        T = (15294.6+1318.8*P+2.4834*P**2)/(8.048+2.8352*np.log(DMg)+2.097*np.log(1.5*C_NM)+2.575*np.log(3*self.df["SiO2_primary_mol"])-1.41*NF+0.222*self.df["H2O_primary_wt"]+0.5*P)
            
        return {'P': P, 'T': T, 'P_err': P * self.T_err/T, 'T_err': self.T_err} 

def compute_sample_pressure_temperature(df, method="PF16", min_SiO2=0.):
    """
    Calculate temperate and pressure of melt equilibration.
    
    Parameters
    ----------
    df : pandas dataframe.
        Dataframe containing the sample primary composition.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().    
    
    method: string
        choice of thermobaric method
        PF16 = Plank & Forsyth (2016), G3
        L09 = Lee et al., (2009), G3
        TGK12_PLG = Till et al., (2012) in plagioclase stability field.
        TGK12_SPL = Till et al., (2012) in spinel stability field.
        G13 = Grove et al., (2013) in garnet stability field.
        BK21_PLG = Brown Krein et al., (2021) in plagioclase stability field.
        BK21_SPL = Brown Krein et al., (2021) in spinel stability field.
        BK21_GNT = Brown Krein et al., (2021) in garnet stability field.
        BK21 = Brown Krein et al., (2021)
        P07_P08 = Putirka et al., (2007) Eq 4. and Putirka (2008) combined.
        SD20 = Sun & Dasgupta (2020), EPSL low SiO2 thermobarometer.
            
    Returns
    -------
    PT  : dict
            The calculated temperature(s) in oC,
            pressures in GPa, and associated errors.  
    """     
    
    if df['SiO2_primary_wt'] < min_SiO2:
        out = {'P': np.nan, 'P_err': np.nan, 'T': np.nan, 'T_err': np.nan}
    elif inspect.isclass(method):
        out = method(df).compute_pressure_temperature()
    elif method == "PF16":
        out = PF16(df).compute_pressure_temperature()
    elif method == "L09":
        out = L09(df).compute_pressure_temperature()
    elif method == "TGK12_PLG":
        out = TGK12_PLG(df).compute_pressure_temperature()
    elif method == "TGK12_SPL":
        out = TGK12_SPL(df).compute_pressure_temperature()
    elif method == "G13":
        out = G13(df).compute_pressure_temperature()
    elif method == "BK21_PLG":
        out = BK21_PLG(df).compute_pressure_temperature()
    elif method == "BK21_SPL":
        out = BK21_SPL(df).compute_pressure_temperature()
    elif method == "BK21_GNT":
        out = BK21_GNT(df).compute_pressure_temperature()
    elif method == "BK21":
        out = BK21(df).compute_pressure_temperature()        
    elif method == "P08":
        out = P08(df).compute_pressure_temperature()        
    elif method == "SD20":
        out = SD20(df).compute_pressure_temperature()        
    else:
        out = thermobar(df).compute_pressure_temperature()
        
    return out


def compute_sample_temperature(df, method="HA15", P=1., min_SiO2=0.):
    """
    Calculate temperate of melt equilibration.
    
    Parameters
    ----------
    df : pandas dataframe.
        Dataframe containing the sample primary composition.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().  
                                    
    method: sring
        choice of thermobaric method
        HA15 = Herzberg & Asimow (2015)
        P07_2 = Putirka et al., (2007) equation 2.
        P07_4 = Putirka et al., (2007) equation 4.
        TGK12_PLG = Till et al., (2012) in plagioclase stability field.
        TGK12_SPL = Till et al., (2012) in spinel stability field.        
        G13 = Grove et al., (2013) in garnet stability field.  
        BK21_PLG = Brown Krein et al., (2021) in plagioclase stability field.
        BK21_SPL = Brown Krein et al., (2021) in spinel stability field.
        BK21_GNT = Brown Krein et al., (2021) in garnet stability field.
        BK21 = Brown Krein et al., (2021).
        SD20 = Sun & Dasgupta (2020), EPSL low SiO2 thermobarometer.
        
    P  : float
        pressure in GPa  
        
    Returns
    -------
    PT  : pandas dataframe
        The calculated temperature(s) in oC,
        assumed pressures in GPa and associated errors.
    """     
    
    if df['SiO2_primary_wt'] < min_SiO2:
        out = {'P': np.nan, 'P_err': np.nan, 'T': np.nan, 'T_err': np.nan}
    elif inspect.isclass(method):
        out = method(df).compute_temperature(P)
    elif method == "HA15":
        out = HA15(df).compute_temperature(P) 
    elif method == "P07_2":
        out = P07_2(df).compute_temperature(P)          
    elif method == "P07_4":
        out = P07_4(df).compute_temperature(P)
    elif method == "B93":
        out = B93(df).compute_temperature(P)        
    elif method == "TGK12_PLG":
        out = TGK12_PLG(df).compute_temperature(P)        
    elif method == "TGK12_SPL":
        out = TGK12_SPL(df).compute_temperature(P)  
    elif method == "G13":
        out = G13(df).compute_temperature(P)      
    elif method == "BK21_PLG":
        out = BK21_PLG(df).compute_temperature(P)        
    elif method == "BK21_SPL":
        out = BK21_SPL(df).compute_temperature(P)  
    elif method == "BK21_GNT":
        out = BK21_GNT(df).compute_temperature(P)  
    elif method == "BK21":
        out = BK21(df).compute_temperature(P)        
    elif method == "SD20":
        out = SD20(df).compute_temperature(P)            
    else:
        out = thermobar(df).compute_temperature(P)
        
    return out

def compute_sample_pressure(df, method="PF16", T=1300., min_SiO2=0.):
    """
    Calculate temperate of melt equilibration
    
    Parameters
    ----------
    df : pandas dataframe.
        Dataframe containing the sample primary composition.
        Should contain only one row. To use with a multi-row dataframe use
        df.apply().  
                                    
    method: sring
        choice of thermobaric method
        PF16 = Plank & Forsyth (2016), G3
        L09 = Lee et al., (2009), G3
        P08 = Putirka (2008) Eq 42.
        
    T  : float
        Temperature in oC
        
    Returns
    -------
    PT  : pandas dataframe
        The calculated temperature(s) in oC,
        assumed pressures in GPa and associated errors.
    """     
    
    if df['SiO2_primary_wt'] < min_SiO2:
        out = {'P': np.nan, 'P_err': np.nan, 'T': np.nan, 'T_err': np.nan}
    elif inspect.isclass(method):
        out = method(df).compute_pressure(T)
    elif method == "PF16":
        out = PF16(df).compute_pressure(T) 
    elif method == "L09":
        out = L09(df).compute_pressure(T)          
    elif method == "P08":
        out = P08(df).compute_pressure(T)        
    else:
        out = thermobar(df).compute_temperature(T)
        
    return out


from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='meltPT',
    version='v1.2.0',
    author = 'F. McNab, P. W. Ball',
    author_email="mcnab@gfz-potsdam.de",
    url="https://github.com/fmcnab/meltPT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    description=("Calculate pressures and temperatures of melting for basaltic rocks and fit with melting models."),
    packages=['meltPT'],
    install_requires=[
        'pandas~=1.3',
        'numpy~=1.21',
        'matplotlib~=3.5',
        'scipy~=1.7',
        'shapely~=1.8',
        'pyMelt~=1.96',
        'pyyaml~=6.0',
        'sympy~=1.10'
    ],
    extras_require={'docs': [
            'sphinx~=5.1',
            'nbsphinx~=0.8',
            'nbsphinx_link~=1.3',
            'sphinx-rtd-theme~=1.0',
            'ipykernel~=6.15'
        ]
    },
    scripts=['meltPT/meltPT'],
    python_requires='>=3.7'
)